from dataclasses import dataclass
from operator import xor
from typing import Optional

SampleCount = int
FrameCount = int


@dataclass(unsafe_hash=True, frozen=True)
class Duration:
    sampling_rate: int
    num_samples: Optional[SampleCount] = None
    num_frames: Optional[FrameCount] = None
    frame_shift: Optional[SampleCount] = None

    def __post_init__(self):
        assert self.sampling_rate is not None, "Sampling rate has to be specified for Duration"
        represents_samples = self.num_samples is not None
        represents_frames = self.represents_frames
        assert xor(represents_samples, represents_frames), \
            "Either num_samples or (num_frames and frame_shift), but not both, need to be specified for Duration"
        if represents_frames:
            object.__setattr__(self, 'num_samples', self.num_frames * self.frame_shift)

    @property
    def represents_frames(self) -> bool:
        return self.frame_shift is not None and self.num_frames is not None

    @property
    def frame_shift_seconds(self) -> float:
        return self.frame_shift / self.sampling_rate

    @property
    def seconds(self) -> float:
        if self.represents_frames:
            return self.num_frames * self.frame_shift / self.sampling_rate
        return self.num_samples / self.sampling_rate

    @property
    def milliseconds(self) -> float:
        return self.seconds * 1000.0

    def __add__(self, other: 'Duration') -> 'Duration':
        assert self.sampling_rate == other.sampling_rate
        if self.represents_frames:
            assert self.frame_shift == other.frame_shift
            return Duration(
                num_frames=self.num_frames + other.num_frames,
                frame_shift=self.frame_shift,
                sampling_rate=self.sampling_rate
            )
        return Duration(
            num_samples=self.num_samples + other.num_samples,
            sampling_rate=self.sampling_rate
        )

    def __sub__(self, other: 'Duration') -> 'Duration':
        return self + (-other)

    def __neg__(self) -> 'Duration':
        if self.represents_frames:
            return Duration(
                num_frames=-self.num_frames,
                frame_shift=self.frame_shift,
                sampling_rate=self.sampling_rate
            )
        return Duration(
            num_samples=-self.num_samples,
            sampling_rate=self.sampling_rate
        )

    def __bool__(self) -> bool:
        return bool(self.num_samples)

    # Ordering and equality operators.
    # We assume that Duration's of different sampling_rate or frame_shift
    # can still be meaningfully compared as time quantities, so we cast them to seconds.

    def __eq__(self, other: 'Duration') -> bool:
        return self.seconds == other.seconds

    def __lt__(self, other: 'Duration') -> bool:
        return self.seconds < other.seconds

    def __gt__(self, other: 'Duration') -> bool:
        return other < self

    def __le__(self, other: 'Duration') -> bool:
        return not (other > self)

    def __ge__(self, other: 'Duration') -> bool:
        return not (self < other)



import click

from typing import List

from lhotse.bin.modes import prepare
from lhotse.recipes import prepare_fisher_english
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument('corpus-dir', type=click.Path(exists=True, file_okay=False))
@click.argument('output-dir', type=click.Path())
@click.option('-ad', '--audio-dirs', type=str, multiple=True, default=['LDC2004T19', 'LDC2005T19'],
              help='Audio dirs, e.g., `LDC2004T19 LDC2005T19`. Multiple corpora can be provided by repeating `-ad`.')
@click.option('-td', '--transcript-dirs', type=str, multiple=True, default=['LDC2004T19', 'LDC2005T19'],
              help='Transcript dirs, e.g., `LDC2004T19 LDC2005T19`. Multiple corpora can be provided by repeating `-ad`.')
@click.option('--absolute-paths', default=False,
              help='Whether to return absolute or relative (to the corpus dir) paths for recordings.')
def fisher_english(
        corpus_dir: Pathlike,
        audio_dirs: List[str],
        transcript_dirs: List[str],
        output_dir: Pathlike,
        absolute_paths: bool
):
    """
    The Fisher English Part 1, 2 corpus preparation.

    \b
    This is conversational telephone speech collected as 2-channel, 8kHz-sampled data. 
    The catalog number LDC2004S13 and LDC2005S13 for audio corpora and LDC2004T19 LDC2005T19 for transcripts.

    This data is not available for free - your institution needs to have an LDC subscription.
    """
    prepare_fisher_english(
        corpus_dir=corpus_dir,
        audio_dirs=audio_dirs,
        transcript_dirs=transcript_dirs,
        output_dir=output_dir,
        absolute_paths=absolute_paths
    )



"""
 Copyright 2019 Johns Hopkins University  (Author: Jesus Villalba)
           2021 Johns Hopkins University  (Author: Piotr Żelasko)
 Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)

This whole module is authored and contributed by Jesus Villalba,
with minor changes by Piotr Żelasko to make it more consistent with Lhotse.

It contains a PyTorch implementation of feature extractors that is very close to Kaldi's
-- notably, it differs in that the preemphasis and DC offset removal are applied in the
time, rather than frequency domain. This should not significantly affect any results, as
confirmed by Jesus.

This implementation works well with autograd and batching, and can be used neural network
layers.
"""
import math
import warnings
from typing import List, Optional

import numpy as np
import torch
from torch import nn

try:
    from torch.fft import rfft as torch_rfft

    _rfft = lambda x: torch_rfft(x, dim=-1)
    _pow_spectrogram = lambda x: x.abs() ** 2
    _spectrogram = lambda x: x.abs()
except ImportError:
    _rfft = lambda x: torch.rfft(x, 1, normalized=False, onesided=True)
    _pow_spectrogram = lambda x: x.pow(2).sum(-1)
    _spectrogram = lambda x: x.pow(2).sum(-1).sqrt()

try:
    from torch.fft import rfft as torch_rfft


    def _rfft(x: torch.Tensor) -> torch.Tensor:
        return torch_rfft(x, dim=-1)


    def _pow_spectrogram(x: torch.Tensor) -> torch.Tensor:
        return x.abs() ** 2


    def _spectrogram(x: torch.Tensor) -> torch.Tensor:
        return x.abs()

except ImportError:

    def _rfft(x: torch.Tensor) -> torch.Tensor:
        return torch.rfft(x, 1, normalized=False, onesided=True)


    def _pow_spectrogram(x: torch.Tensor) -> torch.Tensor:
        return x.pow(2).sum(-1)


    def _spectrogram(x: torch.Tensor) -> torch.Tensor:
        return x.pow(2).sum(-1).sqrt()

from lhotse.utils import EPSILON, Seconds


class Wav2Win(nn.Module):
    def __init__(
            self,
            sampling_rate: int = 16000,
            frame_length: Seconds = 0.025,
            frame_shift: Seconds = 0.01,
            pad_length: Optional[int] = None,
            remove_dc_offset: bool = True,
            preemph_coeff: float = 0.97,
            window_type: str = 'povey',
            dither: float = 0.0,
            snip_edges: bool = False,
            energy_floor: float = EPSILON,
            raw_energy: bool = True,
            return_log_energy: bool = False
    ) -> None:
        super().__init__()
        self.sampling_rate = sampling_rate
        self.frame_length = frame_length
        self.frame_shift = frame_shift
        self.remove_dc_offset = remove_dc_offset
        self.preemph_coeff = preemph_coeff
        self.window_type = window_type
        self.dither = dither
        self.snip_edges = snip_edges
        self.energy_floor = energy_floor
        self.raw_energy = raw_energy
        self.return_log_energy = return_log_energy
        if snip_edges:
            warnings.warn('Setting snip_edges=True is generally incompatible with Lhotse -- '
                          'you might experience mismatched duration/num_frames errors.')

        N = int(math.floor(frame_length * sampling_rate))
        self._length = N
        self._shift = int(math.floor(frame_shift * sampling_rate))

        self._window = nn.Parameter(
            create_frame_window(N, window_type=window_type),
            requires_grad=False)
        self.pad_length = N if pad_length is None else pad_length
        assert self.pad_length >= N, f"pad_length (or fft_length) = {pad_length} cannot be smaller than N = {N}"

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        s = ('{}(sampling_rate={}, frame_length={}, frame_shift={}, pad_length={}, '
             'remove_dc_offset={}, preemph_coeff={}, window_type={} '
             'dither={}, snip_edges={}, energy_floor={}, raw_energy={}, return_log_energy={})').format(
            self.__class__.__name__, self.sampling_rate, self.frame_length, self.frame_shift,
            self.pad_length, self.remove_dc_offset, self.preemph_coeff,
            self.window_type, self.dither, self.snip_edges,
            self.energy_floor, self.raw_energy, self.return_log_energy)
        return s

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Add dither
        if self.dither != 0.0:
            n = torch.randn(x.shape, device=x.device)
            x = x + self.dither * n

        # remove offset
        if self.remove_dc_offset:
            mu = torch.mean(x, dim=1, keepdim=True)
            x = x - mu

        if self.return_log_energy and self.raw_energy:
            # Compute the log energy of each frame
            x_strided = _get_strided_batch(x, self._length, self._shift, self.snip_edges)
            log_energy = _get_log_energy(x_strided, self.energy_floor)  # size (m)

        if self.preemph_coeff != 0.0:
            x_offset = torch.nn.functional.pad(
                x.unsqueeze(1), (1, 0), mode='replicate').squeeze(1)
            x = x - self.preemph_coeff * x_offset[:, :-1]

        x_strided = _get_strided_batch(x, self._length, self._shift, self.snip_edges)

        # Apply window_function to each frame
        x_strided = x_strided * self._window

        # Pad columns with zero until we reach size (batch, num_frames, pad_length)
        if self.pad_length != self._length:
            pad = self.pad_length - self._length
            x_strided = torch.nn.functional.pad(
                x_strided.unsqueeze(1), (0, pad), mode='constant', value=0).squeeze(1)

        if self.return_log_energy:
            return x_strided, log_energy

        return x_strided


class Wav2FFT(nn.Module):
    def __init__(
            self,
            sampling_rate: int = 16000,
            frame_length: Seconds = 0.025,
            frame_shift: Seconds = 0.01,
            fft_length: int = 512,
            remove_dc_offset: bool = True,
            preemph_coeff: float = 0.97,
            window_type: str = 'povey',
            dither: float = 0.0,
            snip_edges: bool = False,
            energy_floor: float = EPSILON,
            raw_energy: bool = True,
            use_energy: bool = True,
    ) -> None:
        super().__init__()

        N = int(math.floor(frame_length * sampling_rate))
        if N > fft_length:
            k = math.ceil(math.log(N) / math.log(2))
            self.fft_length = int(2 ** k)

        self.wav2win = Wav2Win(
            sampling_rate,
            frame_length,
            frame_shift,
            pad_length=fft_length,
            remove_dc_offset=remove_dc_offset,
            preemph_coeff=preemph_coeff,
            window_type=window_type,
            dither=dither,
            snip_edges=snip_edges,
            energy_floor=energy_floor,
            raw_energy=raw_energy,
            return_log_energy=use_energy
        )

        self.fft_length = fft_length
        self.use_energy = use_energy

    @property
    def sampling_rate(self) -> int:
        return self.wav2win.sampling_rate

    @property
    def frame_length(self) -> Seconds:
        return self.wav2win.frame_length

    @property
    def frame_shift(self) -> Seconds:
        return self.wav2win.frame_shift

    @property
    def remove_dc_offset(self) -> bool:
        return self.wav2win.remove_dc_offset

    @property
    def preemph_coeff(self) -> float:
        return self.wav2win.preemph_coeff

    @property
    def window_type(self) -> str:
        return self.wav2win.window_type

    @property
    def dither(self) -> float:
        return self.wav2win.dither

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_strided = self.wav2win(x)
        if self.use_energy:
            x_strided, log_e = x_strided

        X = _rfft(x_strided)

        if self.use_energy:
            X[:, 0, :, 0] = log_e

        return X


class Wav2Spec(Wav2FFT):
    def __init__(
            self,
            sampling_rate: int = 16000,
            frame_length: Seconds = 0.025,
            frame_shift: Seconds = 0.01,
            fft_length: int = 512,
            remove_dc_offset: bool = True,
            preemph_coeff: float = 0.97,
            window_type: str = 'povey',
            dither: float = 0.0,
            snip_edges: bool = False,
            energy_floor: float = EPSILON,
            raw_energy: bool = True,
            use_energy: bool = True,
            use_fft_mag: bool = False,
    ) -> None:
        super().__init__(
            sampling_rate, frame_length, frame_shift, fft_length,
            remove_dc_offset=remove_dc_offset, preemph_coeff=preemph_coeff,
            window_type=window_type, dither=dither, snip_edges=snip_edges,
            energy_floor=energy_floor, raw_energy=raw_energy, use_energy=use_energy)
        self.use_fft_mag = use_fft_mag
        if use_fft_mag:
            self._to_spec = _spectrogram
        else:
            self._to_spec = _pow_spectrogram

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_strided = self.wav2win(x)
        if self.use_energy:
            x_strided, log_e = x_strided

        X = _rfft(x_strided)
        pow_spec = self._to_spec(X)

        if self.use_energy:
            pow_spec[:, 0] = log_e

        return pow_spec


class Wav2LogSpec(Wav2FFT):
    def __init__(
            self,
            sampling_rate: int = 16000,
            frame_length: Seconds = 0.025,
            frame_shift: Seconds = 0.01,
            fft_length: int = 512,
            remove_dc_offset: bool = True,
            preemph_coeff: float = 0.97,
            window_type: str = 'povey',
            dither: float = 0.0,
            snip_edges: bool = False,
            energy_floor: float = EPSILON,
            raw_energy: bool = True,
            use_energy: bool = True,
            use_fft_mag: bool = False,
    ) -> None:
        super().__init__(
            sampling_rate, frame_length, frame_shift, fft_length,
            remove_dc_offset=remove_dc_offset, preemph_coeff=preemph_coeff,
            window_type=window_type, dither=dither, snip_edges=snip_edges,
            energy_floor=energy_floor, raw_energy=raw_energy, use_energy=use_energy)
        self.use_fft_mag = use_fft_mag
        if use_fft_mag:
            self._to_spec = _spectrogram
        else:
            self._to_spec = _pow_spectrogram

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_strided = self.wav2win(x)
        if self.use_energy:
            x_strided, log_e = x_strided

        X = _rfft(x_strided)
        pow_spec = self._to_spec(X)

        pow_spec = (pow_spec + 1e-15).log()

        if self.use_energy:
            pow_spec[:, 0] = log_e

        return pow_spec


class Wav2LogFilterBank(Wav2FFT):
    def __init__(
            self,
            sampling_rate: int = 16000,
            frame_length: Seconds = 0.025,
            frame_shift: Seconds = 0.01,
            fft_length: int = 512,
            remove_dc_offset: bool = True,
            preemph_coeff: float = 0.97,
            window_type: str = 'povey',
            dither: float = 0.0,
            snip_edges: bool = False,
            energy_floor: float = EPSILON,
            raw_energy: bool = True,
            use_energy: bool = False,
            use_fft_mag: bool = False,
            low_freq: float = 20.0,
            high_freq: float = -400.0,
            num_filters: int = 80,
            norm_filters: bool = False,
    ):

        super().__init__(
            sampling_rate, frame_length, frame_shift, fft_length,
            remove_dc_offset=remove_dc_offset, preemph_coeff=preemph_coeff,
            window_type=window_type, dither=dither, snip_edges=snip_edges,
            energy_floor=energy_floor, raw_energy=raw_energy, use_energy=use_energy)

        self.use_fft_mag = use_fft_mag
        self.low_freq = low_freq
        self.high_freq = high_freq
        self.num_filters = num_filters
        self.norm_filters = norm_filters

        if use_fft_mag:
            self._to_spec = _spectrogram
        else:
            self._to_spec = _pow_spectrogram

        fb = create_mel_scale(
            num_filters=num_filters,
            fft_length=fft_length,
            sampling_rate=sampling_rate,
            low_freq=low_freq,
            high_freq=high_freq,
            norm_filters=norm_filters
        )
        self._fb = nn.Parameter(
            torch.tensor(fb, dtype=torch.get_default_dtype()),
            requires_grad=False)

    def forward(self, x):
        x_strided = self.wav2win(x)
        if self.use_energy:
            x_strided, log_e = x_strided

        X = _rfft(x_strided)
        pow_spec = self._to_spec(X)

        try:
            from torch.cuda.amp import autocast
        except ImportError:
            warnings.warn('Could not import torch.cuda.amp.autocast -- '
                          'when using mixed precision with another package such as apex, '
                          'you might experience numerical stability issues.')
            pow_spec = torch.matmul(pow_spec.float(), self._fb.float())
        else:
            with autocast(enabled=False):
                pow_spec = torch.matmul(pow_spec.float(), self._fb.float())
        pow_spec = (pow_spec + 1e-10).log()
        if self.use_energy:
            pow_spec = torch.cat((log_e.unsqueeze(-1), pow_spec), dim=-1)

        return pow_spec


class Wav2MFCC(Wav2FFT):
    def __init__(
            self,
            sampling_rate: int = 16000,
            frame_length: Seconds = 0.025,
            frame_shift: Seconds = 0.01,
            fft_length: int = 512,
            remove_dc_offset: bool = True,
            preemph_coeff: float = 0.97,
            window_type: str = 'povey',
            dither: float = 0.0,
            snip_edges: bool = False,
            energy_floor: float = EPSILON,
            raw_energy: bool = True,
            use_energy: bool = False,
            use_fft_mag: bool = False,
            low_freq: float = 20.0,
            high_freq: float = -400.0,
            num_filters: int = 23,
            norm_filters: bool = False,
            num_ceps: int = 13,
            cepstral_lifter: int = 22,
    ):

        super().__init__(
            sampling_rate, frame_length, frame_shift, fft_length,
            remove_dc_offset=remove_dc_offset, preemph_coeff=preemph_coeff,
            window_type=window_type, dither=dither, snip_edges=snip_edges,
            energy_floor=energy_floor, raw_energy=raw_energy, use_energy=use_energy)

        self.use_fft_mag = use_fft_mag
        self.low_freq = low_freq
        self.high_freq = high_freq
        self.num_filters = num_filters
        self.norm_filters = norm_filters
        self.num_ceps = num_ceps
        self.cepstral_lifter = cepstral_lifter

        if use_fft_mag:
            self._to_spec = _spectrogram
        else:
            self._to_spec = _pow_spectrogram

        fb = create_mel_scale(
            num_filters=num_filters,
            fft_length=fft_length,
            sampling_rate=sampling_rate,
            low_freq=low_freq,
            high_freq=high_freq,
            norm_filters=norm_filters
        )
        self._fb = nn.Parameter(
            torch.tensor(fb, dtype=torch.get_default_dtype()),
            requires_grad=False)
        self._dct = nn.Parameter(
            self.make_dct_matrix(self.num_ceps, self.num_filters),
            requires_grad=False)
        self._lifter = nn.Parameter(
            self.make_lifter(self.num_ceps, self.cepstral_lifter),
            requires_grad=False)

    @staticmethod
    def make_lifter(N, Q):
        """Makes the liftering function

           Args:
             N: Number of cepstral coefficients.
             Q: Liftering parameter
           Returns:
             Liftering vector.
        """
        if Q == 0:
            return 1
        return 1 + 0.5 * Q * torch.sin(math.pi * torch.arange(
            N, dtype=torch.get_default_dtype()) / Q)

    @staticmethod
    def make_dct_matrix(num_ceps, num_filters):
        n = torch.arange(float(num_filters)).unsqueeze(1)
        k = torch.arange(float(num_ceps))
        dct = torch.cos(math.pi / float(num_filters) * (n + 0.5) * k)  # size (n_mfcc, n_mels)
        dct[:, 0] *= 1.0 / math.sqrt(2.0)
        dct *= math.sqrt(2.0 / float(num_filters))
        return dct

    def forward(self, x):
        x_strided = self.wav2win(x)
        if self.use_energy:
            x_strided, log_e = x_strided

        X = _rfft(x_strided)
        pow_spec = self._to_spec(X)

        try:
            from torch.cuda.amp import autocast
        except ImportError:
            warnings.warn('Could not import torch.cuda.amp.autocast -- '
                          'when using mixed precision with another package such as apex, '
                          'you might experience numerical stability issues.')
            pow_spec = torch.matmul(pow_spec.float(), self._fb.float())
        else:
            with autocast(enabled=False):
                pow_spec = torch.matmul(pow_spec.float(), self._fb.float())

        pow_spec = (pow_spec + 1e-10).log()

        mfcc = torch.matmul(pow_spec, self._dct)
        if self.cepstral_lifter > 0:
            mfcc *= self._lifter

        if self.use_energy:
            mfcc[:, 0] = log_e

        return mfcc


def _get_strided_batch(waveform, window_length, window_shift, snip_edges):
    r"""Given a waveform (1D tensor of size ``num_samples``), it returns a 2D tensor (m, ``window_size``)
    representing how the window is shifted along the waveform. Each row is a frame.
    Args:
        waveform (torch.Tensor): Tensor of size ``num_samples``
        window_size (int): Frame length
        window_shift (int): Frame shift
        snip_edges (bool): If True, end effects will be handled by outputting only frames that completely fit
            in the file, and the number of frames depends on the frame_length.  If False, the number of frames
            depends only on the frame_shift, and we reflect the data at the ends.
    Returns:
        torch.Tensor: 3D tensor of size (m, ``window_size``) where each row is a frame
    """
    assert waveform.dim() == 2
    batch_size = waveform.size(0)
    num_samples = waveform.size(-1)

    if snip_edges:
        if num_samples < window_length:
            return torch.empty((0, 0, 0))
        else:
            num_frames = 1 + (num_samples - window_length) // window_shift
    else:
        num_frames = (num_samples + (window_shift // 2)) // window_shift
        new_num_samples = (num_frames - 1) * window_shift + window_length
        npad = new_num_samples - num_samples
        npad_left = int((window_length - window_shift) // 2)
        npad_right = npad - npad_left
        # waveform = nn.functional.pad(waveform, (npad_left, npad_right), mode='reflect')
        pad_left = torch.flip(waveform[:, 1:npad_left + 1], (1,))
        pad_right = torch.flip(waveform[:, -npad_right - 1:-1], (1,))
        waveform = torch.cat((pad_left, waveform, pad_right), dim=1)

    strides = (waveform.stride(0), window_shift * waveform.stride(1), waveform.stride(1))
    sizes = (batch_size, num_frames, window_length)
    return waveform.as_strided(sizes, strides)


def _get_log_energy(x: torch.Tensor, energy_floor: float) -> torch.Tensor:
    """
    Returns the log energy of size (m) for a strided_input (m,*)
    """
    log_energy = (x.pow(2).sum(-1) + 1e-15).log()  # size (m)
    if energy_floor > 0.0:
        log_energy = torch.max(
            log_energy,
            torch.tensor(math.log(energy_floor), dtype=torch.get_default_dtype())
        )

    return log_energy


def create_mel_scale(
        num_filters: int,
        fft_length: int,
        sampling_rate: int,
        low_freq: float = 0,
        high_freq: Optional[float] = None,
        norm_filters: bool = True,
):
    if high_freq is None or high_freq == 0:
        high_freq = sampling_rate / 2
    if high_freq < 0:
        high_freq = sampling_rate / 2 + high_freq

    mel_low_freq = lin2mel(low_freq)
    mel_high_freq = lin2mel(high_freq)
    melfc = np.linspace(mel_low_freq, mel_high_freq, num_filters + 2)
    mels = lin2mel(np.linspace(0, sampling_rate, fft_length))

    B = np.zeros((int(fft_length / 2 + 1), num_filters), dtype=np.float32)
    for k in range(num_filters):
        left_mel = melfc[k]
        center_mel = melfc[k + 1]
        right_mel = melfc[k + 2]
        for j in range(int(fft_length / 2)):
            mel_j = mels[j]
            if left_mel < mel_j < right_mel:
                if mel_j <= center_mel:
                    B[j, k] = (mel_j - left_mel) / (center_mel - left_mel)
                else:
                    B[j, k] = (right_mel - mel_j) / (right_mel - center_mel)

    if norm_filters:
        B = B / np.sum(B, axis=0, keepdims=True)

    return B


def available_windows() -> List[str]:
    return [HAMMING, HANNING, POVEY, RECTANGULAR, BLACKMAN]


HAMMING = 'hamming'
HANNING = 'hanning'
POVEY = 'povey'
RECTANGULAR = 'rectangular'
BLACKMAN = 'blackman'


def create_frame_window(window_size, window_type: str = 'povey', blackman_coeff=0.42):
    r"""Returns a window function with the given type and size
    """
    if window_type == HANNING:
        return torch.hann_window(window_size, periodic=True)
    elif window_type == HAMMING:
        return torch.hamming_window(window_size, periodic=True, alpha=0.54, beta=0.46)
    elif window_type == POVEY:
        a = 2 * math.pi / window_size
        window_function = torch.arange(window_size, dtype=torch.get_default_dtype())
        return (0.5 - 0.5 * torch.cos(a * window_function)).pow(0.85)
    elif window_type == RECTANGULAR:
        return torch.ones(window_size, dtype=torch.get_default_dtype())
    elif window_type == BLACKMAN:
        a = 2 * math.pi / window_size
        window_function = torch.arange(window_size, dtype=torch.get_default_dtype())
        return blackman_coeff - 0.5 * torch.cos(a * window_function) + \
               (0.5 - blackman_coeff) * torch.cos(2 * a * window_function)
    else:
        raise Exception(f'Invalid window type: {window_type}')


def lin2mel(x):
    return 1127.0 * np.log(1 + x / 700)


def mel2lin(x):
    return 700 * (np.exp(x / 1127.0) - 1)



import bisect
import random
from typing import Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import torch

from lhotse import CutSet
from lhotse.utils import Pathlike

__all__ = [
    'GlobalMVN',
    'SpecAugment',
    'RandomizedSmoothing'
]


class GlobalMVN(torch.nn.Module):
    """Apply global mean and variance normalization"""

    def __init__(self, feature_dim: int):
        super().__init__()
        self.feature_dim = feature_dim
        self.register_buffer("norm_means", torch.zeros(feature_dim))
        self.register_buffer("norm_stds", torch.ones(feature_dim))

    @classmethod
    def from_cuts(cls, cuts: CutSet, max_cuts: Optional[int] = None) -> "GlobalMVN":
        stats = cuts.compute_global_feature_stats(max_cuts=max_cuts)
        stats = {name: torch.as_tensor(value) for name, value in stats.items()}
        feature_dim, = stats["norm_means"].shape
        global_mvn = cls(feature_dim)
        global_mvn.load_state_dict(stats)
        return global_mvn

    @classmethod
    def from_file(cls, stats_file: Pathlike) -> "GlobalMVN":
        stats = torch.load(stats_file)
        feature_dim, = stats["norm_means"].shape
        global_mvn = cls(feature_dim)
        global_mvn.load_state_dict(stats)
        return global_mvn

    def to_file(self, stats_file: Pathlike):
        torch.save(self.state_dict(), stats_file)

    def forward(self, features: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        return (features - self.norm_means) / self.norm_stds

    def inverse(self, features: torch.Tensor) -> torch.Tensor:
        return features * self.norm_stds + self.norm_means


class RandomizedSmoothing(torch.nn.Module):
    """
    Randomized smoothing - gaussian noise added to an input waveform, or a batch of waveforms.
    The summed audio is clipped to ``[-1.0, 1.0]`` before returning.
    """

    def __init__(
            self,
            sigma: Union[float, Sequence[Tuple[int, float]]] = 0.1,
            sample_sigma: bool = True,
            p: float = 0.3,
    ):
        """
        RandomizedSmoothing's constructor.

        :param sigma: standard deviation of the gaussian noise. Either a constant float, or a schedule,
            i.e. a list of tuples that specify which value to use from which step.
            For example, ``[(0, 0.01), (1000, 0.1)]`` means that from steps 0-999, the sigma value
            will be 0.01, and from step 1000 onwards, it will be 0.1.
        :param sample_sigma: when ``False``, then sigma is used as the standard deviation in each forward step.
            When ``True``, the standard deviation is sampled from a uniform distribution of
            ``[-sigma, sigma]`` for each forward step.
        :param p: the probability of applying this transform.
        """
        super().__init__()
        self.sigma = sigma
        self.sample_sigma = sample_sigma
        self.p = p
        self.step = 0

    def forward(self, audio: torch.Tensor, *args, **kwargs) -> torch.Tensor:

        # Determine the stddev value
        if isinstance(self.sigma, float):
            # Use a constant stddev value
            sigma = self.sigma
        else:
            # Determine the right stddev value from a given schedule.
            sigma = schedule_value_for_step(self.sigma, self.step)
            self.step += 1

        if self.sample_sigma:
            # In this mode stddev is stochastic itself
            # and is sampled from uniform distribution bounded by [-sigma, sigma] .
            mask_shape = (audio.shape[0],) + tuple(1 for _ in audio.shape[1:])
            # Sigma is of shape (batch_size, 1) - different for each noise example.
            sigma = sigma * (2 * torch.rand(mask_shape) - 1)

        # Create the random noise examples with identical sigma's.
        noise = sigma * torch.randn_like(audio)

        # Apply the transform with a probability p -> mask noise examples with probability 1 - p.
        noise_mask = random_mask_along_batch_axis(noise, p=1.0 - self.p)
        noise = noise * noise_mask

        return torch.clip(audio + noise, min=-1.0, max=1.0)


class SpecAugment(torch.nn.Module):
    """
    SpecAugment performs three augmentations:
    - time warping of the feature matrix
    - masking of ranges of features (frequency bands)
    - masking of ranges of frames (time)

    The current implementation works with batches, but processes each example separately
    in a loop rather than simultaneously to achieve different augmentation parameters for
    each example.
    """

    def __init__(
            self,
            time_warp_factor: Optional[int] = 80,
            num_feature_masks: int = 1,
            features_mask_size: int = 13,
            num_frame_masks: int = 1,
            frames_mask_size: int = 70,
            max_frames_mask_fraction: float = 0.2,
            p=0.5,
    ):
        """
        SpecAugment's constructor.

        :param time_warp_factor: parameter for the time warping; larger values mean more warping.
            Set to ``None``, or less than ``1``, to disable.
        :param num_feature_masks: how many feature masks should be applied. Set to ``0`` to disable.
        :param features_mask_size: the width of the feature mask (expressed in the number of masked feature bins).
            This is the ``T`` parameter from the SpecAugment paper.
        :param num_frame_masks: how many frame (temporal) masks should be applied. Set to ``0`` to disable.
        :param frames_mask_size: the width of the frame (temporal) masks (expressed in the number of masked frames).
            This is the ``F`` parameter from the SpecAugment paper.
        :param max_frames_mask_fraction: limits the size of the frame (temporal) mask to this value times the length
            of the utterance (or supervision segment).
            This is the parameter denoted by ``p`` in the SpecAugment paper.
        :param p: the probability of applying this transform.
            It is different from ``p`` in the SpecAugment paper!
        """
        super().__init__()
        assert 0 <= p <= 1
        assert num_feature_masks >= 0
        assert num_frame_masks >= 0
        assert features_mask_size > 0
        assert frames_mask_size > 0
        self.time_warp_factor = time_warp_factor
        self.num_feature_masks = num_feature_masks
        self.features_mask_size = features_mask_size
        self.num_frame_masks = num_frame_masks
        self.frames_mask_size = frames_mask_size
        self.max_frames_mask_fraction = max_frames_mask_fraction
        self.p = p

    def forward(
            self,
            features: torch.Tensor,
            supervision_segments: Optional[torch.IntTensor] = None,
            *args, **kwargs
    ) -> torch.Tensor:
        """
        Computes SpecAugment for a batch of feature matrices.

        Since the batch will usually already be padded, the user can optionally
        provide a ``supervision_segments`` tensor that will be used to apply SpecAugment
        only to selected areas of the input. The format of this input is described below.

        :param features: a batch of feature matrices with shape ``(B, T, F)``.
        :param supervision_segments: an int tensor of shape ``(S, 3)``. ``S`` is the number of
            supervision segments that exist in ``features`` -- there may be either
            less or more than the batch size.
            The second dimension encoder three kinds of information:
            the sequence index of the corresponding feature matrix in `features`,
            the start frame index, and the number of frames for each segment.
        :return: a tensor of shape ``(T, F)``, or a batch of them with shape ``(B, T, F)``
        """
        assert len(features.shape) == 3, 'SpecAugment only supports batches of ' \
                                         'single-channel feature matrices.'
        features = features.clone()
        if supervision_segments is None:
            # No supervisions - apply spec augment to full feature matrices.
            for sequence_idx in range(features.size(0)):
                features[sequence_idx] = self._forward_single(features[sequence_idx])
        else:
            # Supervisions provided - we will apply time warping only on the supervised areas.
            for sequence_idx, start_frame, num_frames in supervision_segments:
                end_frame = start_frame + num_frames
                features[sequence_idx, start_frame: end_frame] = self._forward_single(
                    features[sequence_idx, start_frame: end_frame],
                    warp=True,
                    mask=False
                )
            # ... and then time-mask the full feature matrices. Note that in this mode,
            # it might happen that masks are applied to different sequences/examples
            # than the time warping.
            for sequence_idx in range(features.size(0)):
                features[sequence_idx] = self._forward_single(
                    features[sequence_idx],
                    warp=False,
                    mask=True
                )
        return features

    def _forward_single(self, features: torch.Tensor, warp: bool = True, mask: bool = True) -> torch.Tensor:
        """
        Apply SpecAugment to a single feature matrix of shape (T, F).
        """
        if random.random() > self.p:
            # Randomly choose whether this transform is applied
            return features
        if warp:
            if self.time_warp_factor is not None and self.time_warp_factor >= 1:
                features = time_warp(features, factor=self.time_warp_factor)
        if mask:
            from torchaudio.functional import mask_along_axis
            mean = features.mean()
            for _ in range(self.num_feature_masks):
                features = mask_along_axis(
                    features.unsqueeze(0),
                    mask_param=self.features_mask_size,
                    mask_value=mean,
                    axis=2
                ).squeeze(0)
            for _ in range(self.num_frame_masks):
                max_mask_frames = min(self.frames_mask_size, self.max_frames_mask_fraction * features.size(0))
                features = mask_along_axis(
                    features.unsqueeze(0),
                    mask_param=max_mask_frames,
                    mask_value=mean,
                    axis=1
                ).squeeze(0)
        return features


def time_warp(features: torch.Tensor, factor: int) -> torch.Tensor:
    """
    Time warping as described in the SpecAugment paper.
    Implementation based on Espresso:
    https://github.com/freewym/espresso/blob/master/espresso/tools/specaug_interpolate.py#L51

    :param features: input tensor of shape ``(T, F)``
    :param factor: time warping parameter.
    :return: a warped tensor of shape ``(T, F)``
    """
    t = features.size(0)
    if t - factor <= factor + 1:
        return features
    center = np.random.randint(factor + 1, t - factor)
    warped = np.random.randint(center - factor, center + factor + 1)
    if warped == center:
        return features
    features = features.unsqueeze(0).unsqueeze(0)
    left = torch.nn.functional.interpolate(
        features[:, :, :center, :], size=(warped, features.size(3)),
        mode="bicubic", align_corners=False,
    )
    right = torch.nn.functional.interpolate(
        features[:, :, center:, :], size=(t - warped, features.size(3)),
        mode="bicubic", align_corners=False,
    )
    return torch.cat((left, right), dim=2).squeeze(0).squeeze(0)


T = TypeVar('T')


def schedule_value_for_step(schedule: Sequence[Tuple[int, T]], step: int) -> T:
    milestones, values = zip(*schedule)
    assert milestones[0] <= step, f"Cannot determine the scheduled value for step {step} with schedule: {schedule}. " \
                                  f"Did you forget to add the first part of the schedule " \
                                  f"for steps below {milestones[0]}?"
    idx = bisect.bisect_right(milestones, step) - 1
    return values[idx]


def random_mask_along_batch_axis(tensor: torch.Tensor, p: float = 0.5) -> torch.Tensor:
    """
    For a given tensor with shape (N, d1, d2, d3, ...), returns a mask with shape (N, 1, 1, 1, ...),
    that randomly masks the samples in a batch.

    E.g. for a 2D input matrix it looks like:

        >>> [[0., 0., 0., ...],
        ...  [1., 1., 1., ...],
        ...  [0., 0., 0., ...]]

    :param tensor: the input tensor.
    :param p: the probability of masking an element.
    """
    mask_shape = (tensor.shape[0],) + tuple(1 for _ in tensor.shape[1:])
    mask = (torch.rand(mask_shape) > p).to(torch.float32)
    return mask



import click

from lhotse.bin.modes import prepare
from lhotse.recipes.nsc import NSC_PARTS, prepare_nsc
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument('corpus_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
@click.option('-p', '--dataset-part', type=click.Choice(NSC_PARTS), default='PART3_SameCloseMic',
              help='Which part of NSC should be prepared')
def nsc(
        corpus_dir: Pathlike,
        output_dir: Pathlike,
        dataset_part: str
):
    """
    This is a data preparation recipe for the National Corpus of Speech in Singaporean English.
    """
    prepare_nsc(corpus_dir, dataset_part=dataset_part, output_dir=output_dir)



from .base import (
    FeatureExtractor,
    FeatureSet,
    FeatureSetBuilder,
    Features,
    create_default_feature_extractor,
)
from .fbank import (
    Fbank,
    FbankConfig
)
from .io import (
    ChunkedLilcomHdf5Reader,
    ChunkedLilcomHdf5Writer,
    FeaturesReader,
    FeaturesWriter,
    KaldiReader,
    LilcomFilesReader,
    LilcomFilesWriter,
    LilcomHdf5Reader,
    LilcomHdf5Writer,
    LilcomURLReader,
    LilcomURLWriter,
    NumpyFilesReader,
    NumpyFilesWriter,
    NumpyHdf5Reader,
    NumpyHdf5Writer,
    available_storage_backends,
    close_cached_file_handles
)
from .kaldi.extractors import (
    KaldiFbank,
    KaldiFbankConfig,
    KaldiMfcc,
    KaldiMfccConfig
)
from .librosa_fbank import (
    LibrosaFbank,
    LibrosaFbankConfig
)
from .mfcc import (
    Mfcc,
    MfccConfig
)
from .mixer import (
    FeatureMixer
)
from .spectrogram import (
    Spectrogram,
    SpectrogramConfig
)



import warnings
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

from lhotse import CutSet, FeatureSet, Features, Seconds
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.audio import audioread_info
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, compute_num_samples, is_module_available


def get_duration(
        path: Pathlike,
) -> float:
    """
    Read a audio file, it supports pipeline style wave path and real waveform.

    :param path: Path to an audio file or a Kaldi-style pipe.
    :return: float duration of the recording, in seconds.
    """
    path = str(path)
    if path.strip().endswith('|'):
        if not is_module_available('kaldiio'):
            raise ValueError("To read Kaldi's data dir where wav.scp has 'pipe' inputs, "
                             "please 'pip install kaldiio' first.")
        from kaldiio import load_mat
        # Note: kaldiio.load_mat returns (sampling_rate: int, samples: 1-D np.array[int])
        sampling_rate, samples = load_mat(path)
        assert len(samples.shape) == 1
        duration = samples.shape[0] / sampling_rate
        return duration
    try:
        # Try to parse the file using pysoundfile first.
        import soundfile
        info = soundfile.info(path)
    except:
        # Try to parse the file using audioread as a fallback.
        info = audioread_info(path)
    return info.duration


def load_kaldi_data_dir(
        path: Pathlike,
        sampling_rate: int,
        frame_shift: Optional[Seconds] = None,
) -> Tuple[RecordingSet, Optional[SupervisionSet], Optional[FeatureSet]]:
    """
    Load a Kaldi data directory and convert it to a Lhotse RecordingSet and SupervisionSet manifests.
    For this to work, at least the wav.scp file must exist.
    SupervisionSet is created only when a segments file exists.
    All the other files (text, utt2spk, etc.) are optional, and some of them might not be handled yet.
    In particular, feats.scp files are ignored.
    """
    path = Path(path)
    assert path.is_dir()

    # must exist for RecordingSet
    recordings = load_kaldi_text_mapping(path / 'wav.scp', must_exist=True)

    durations = {}
    for recording_id, path_or_cmd in recordings.items():
        duration = get_duration(path_or_cmd)
        durations[recording_id] = duration

    recording_set = RecordingSet.from_recordings(
        Recording(
            id=recording_id,
            sources=[
                AudioSource(
                    type='command' if path_or_cmd.endswith('|') else 'file',
                    channels=[0],
                    source=path_or_cmd[:-1] if path_or_cmd.endswith('|') else path_or_cmd
                )
            ],
            sampling_rate=sampling_rate,
            num_samples=compute_num_samples(durations[recording_id], sampling_rate),
            duration=durations[recording_id]
        )
        for recording_id, path_or_cmd in recordings.items()
    )

    supervision_set = None
    segments = path / 'segments'
    if segments.is_file():
        with segments.open() as f:
            supervision_segments = [l.strip().split() for l in f]

        texts = load_kaldi_text_mapping(path / 'text')
        speakers = load_kaldi_text_mapping(path / 'utt2spk')
        genders = load_kaldi_text_mapping(path / 'spk2gender')
        languages = load_kaldi_text_mapping(path / 'utt2lang')

        supervision_set = SupervisionSet.from_segments(
            SupervisionSegment(
                id=segment_id,
                recording_id=recording_id,
                start=float(start),
                duration=float(end) - float(start),
                channel=0,
                text=texts[segment_id],
                language=languages[segment_id],
                speaker=speakers[segment_id],
                gender=genders[speakers[segment_id]]
            )
            for segment_id, recording_id, start, end in supervision_segments
        )

    feature_set = None
    feats_scp = path / 'feats.scp'
    if feats_scp.exists() and is_module_available('kaldiio'):
        if frame_shift is not None:
            import kaldiio
            from lhotse.features.io import KaldiReader
            feature_set = FeatureSet.from_features(
                Features(
                    type='kaldiio',
                    num_frames=mat.shape[0],
                    num_features=mat.shape[1],
                    frame_shift=frame_shift,
                    sampling_rate=sampling_rate,
                    start=0,
                    duration=mat.shape[0] * frame_shift,
                    storage_type=KaldiReader.name,
                    storage_path=str(feats_scp),
                    storage_key=utt_id,
                    recording_id=supervision_set[utt_id].recording_id if supervision_set is not None else utt_id,
                    channels=0
                ) for utt_id, mat in kaldiio.load_scp_sequential(str(feats_scp))
            )
        else:
            warnings.warn(f"Failed to import Kaldi 'feats.scp' to Lhotse: "
                          f"frame_shift must be not None. "
                          f"Feature import omitted.")

    return recording_set, supervision_set, feature_set


def export_to_kaldi(
        recordings: RecordingSet,
        supervisions: SupervisionSet,
        output_dir: Pathlike
):
    """
    Export a pair of ``RecordingSet`` and ``SupervisionSet`` to a Kaldi data directory.
    Currently, it only supports single-channel recordings that have a single ``AudioSource``.

    The ``RecordingSet`` and ``SupervisionSet`` must be compatible, i.e. it must be possible to create a
    ``CutSet`` out of them.

    :param recordings: a ``RecordingSet`` manifest.
    :param supervisions: a ``SupervisionSet`` manifest.
    :param output_dir: path where the Kaldi-style data directory will be created.
    """
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    assert all(len(r.sources) == 1 for r in recordings), "Kaldi export of Recordings with multiple audio sources " \
                                                         "is currently not supported."
    assert all(r.num_channels == 1 for r in recordings), "Kaldi export of multi-channel Recordings is currently " \
                                                         "not supported."

    # Create a simple CutSet that ties together the recording <-> supervision information.
    cuts = CutSet.from_manifests(recordings=recordings, supervisions=supervisions).trim_to_supervisions()

    # wav.scp
    save_kaldi_text_mapping(
        data={
            recording.id: make_wavscp_string(source, sampling_rate=recording.sampling_rate)
            for recording in recordings
            for source in recording.sources
        },
        path=output_dir / 'wav.scp'
    )
    # segments
    save_kaldi_text_mapping(
        data={cut.supervisions[0].id: f'{cut.recording_id} {cut.start} {cut.end}' for cut in cuts},
        path=output_dir / 'segments'
    )
    # text
    save_kaldi_text_mapping(
        data={cut.supervisions[0].id: cut.supervisions[0].text for cut in cuts},
        path=output_dir / 'text'
    )
    # utt2spk
    save_kaldi_text_mapping(
        data={cut.supervisions[0].id: cut.supervisions[0].speaker for cut in cuts},
        path=output_dir / 'utt2spk'
    )
    # utt2dur
    save_kaldi_text_mapping(
        data={cut.supervisions[0].id: cut.duration for cut in cuts},
        path=output_dir / 'utt2dur'
    )
    # reco2dur
    save_kaldi_text_mapping(
        data={recording.id: recording.duration for recording in recordings},
        path=output_dir / 'reco2dur'
    )
    # utt2lang [optional]
    if all(s.language is not None for s in supervisions):
        save_kaldi_text_mapping(
            data={cut.supervisions[0].id: cut.supervisions[0].language for cut in cuts},
            path=output_dir / 'utt2lang'
        )
    # utt2gender [optional]
    if all(s.gender is not None for s in supervisions):
        save_kaldi_text_mapping(
            data={cut.supervisions[0].id: cut.supervisions[0].gender for cut in cuts},
            path=output_dir / 'utt2gender'
        )


def load_kaldi_text_mapping(path: Path, must_exist: bool = False) -> Dict[str, Optional[str]]:
    """Load Kaldi files such as utt2spk, spk2gender, text, etc. as a dict."""
    mapping = defaultdict(lambda: None)
    if path.is_file():
        with path.open() as f:
            mapping = dict(line.strip().split(maxsplit=1) for line in f)
    elif must_exist:
        raise ValueError(f"No such file: {path}")
    return mapping


def save_kaldi_text_mapping(data: Dict[str, Any], path: Path):
    """Save flat dicts to Kaldi files such as utt2spk, spk2gender, text, etc."""
    with path.open('w') as f:
        for key, value in data.items():
            print(key, value, file=f)


def make_wavscp_string(source: AudioSource, sampling_rate: int) -> str:
    if source.type == 'url':
        raise ValueError("URL audio sources are not supported by Kaldi.")
    elif source.type == 'command':
        return f'{source.source} |'
    elif source.type == 'file':
        if Path(source.source).suffix == '.wav':
            return source.source
        else:
            return f'ffmpeg -i {source.source} -ar {sampling_rate} -f wav pipe:1 |'
    else:
        raise ValueError(f"Unknown AudioSource type: {source.type}")



from functools import reduce
from operator import add
from typing import Callable, Optional

from lhotse import CutSet
from lhotse.cut import Cut
from lhotse.dataset.sampling.base import CutSampler


class ZipSampler(CutSampler):
    """
    :class:`.ZipSampler` takes several samplers as input and concatenates their
    sampled batch cuts together into a single list.
    It is helpful for ensuring that each batch consists of some proportion of cuts
    coming from different sources.

    The input samplers do not have to provide the same number of batches -- when
    any of the samplers becomes depleted, the iteration will stop (like with
    Python's ``zip()`` function).

    Example::

        >>> sampler = ZipSampler(
        ...     SingleCutSampler(cuts_corpusA, max_duration=250, shuffle=True),
        ...     SingleCutSampler(cuts_corpusB, max_duration=100, shuffle=True),
        ... )
        >>> for cut in sampler:
        ...     pass  # profit
    """

    def __init__(self, *samplers: CutSampler) -> None:
        super().__init__()
        self.samplers = samplers

    @property
    def remaining_duration(self) -> Optional[float]:
        """
        Remaining duration of data left in the sampler (may be inexact due to float arithmetic).

        .. note: For ZipSampler, it's the minimum of remaining durations in its sub-samplers.
        """
        try:
            return min(s.remaining_duration for s in self.samplers)
        except TypeError:
            return None

    @property
    def remaining_cuts(self) -> Optional[int]:
        """
        Remaining number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).

        .. note: For ZipSampler, it's the minimum of remaining cuts in its sub-samplers.
        """
        try:
            return min(s.remaining_cuts for s in self.samplers)
        except TypeError:
            return None

    @property
    def num_cuts(self) -> Optional[int]:
        """
        Total number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).

        .. note: For ZipSampler, it's the minimum of num cuts in its sub-samplers.
        """
        try:
            return min(s.num_cuts for s in self.samplers)
        except TypeError:
            return None

    def __iter__(self):
        for sampler in self.samplers:
            iter(sampler)
        return self

    def _next_batch(self) -> CutSet:
        cuts = []
        for sampler in self.samplers:
            cuts.extend(next(sampler))
        return CutSet.from_cuts(cuts)

    def set_epoch(self, epoch: int) -> None:
        """
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        :param epoch: Epoch number.
        """
        for s in self.samplers:
            s.set_epoch(epoch)
        super().set_epoch(epoch)

    def filter(self, predicate: Callable[[Cut], bool]) -> None:
        """
        Add a constraint on individual cuts that has to be satisfied to consider them.

        Can be useful when handling large, lazy manifests where it is not feasible to
        pre-filter them before instantiating the sampler.

        When set, we will remove the ``__len__`` attribute on the sampler, as it is now
        determined dynamically.

        Example:
            >>> cuts = CutSet(...)
            ... sampler = SingleCutSampler(cuts, max_duration=100.0)
            ... # Retain only the cuts that have at least 1s and at most 20s duration.
            ... sampler.filter(lambda cut: 1.0 <= cut.duration <= 20.0)
        """
        for sampler in self.samplers:
            sampler.filter(predicate)

    def __len__(self):
        if self.num_batches is None:
            self.num_batches = min(len(sampler) for sampler in self.samplers)
        return self.num_batches

    def get_report(self) -> str:
        """Returns a string describing the statistics of the sampling process so far."""
        total_diagnostics = reduce(
            add, (sampler.diagnostics for sampler in self.samplers)
        )
        return total_diagnostics.get_report()



from .bucketing import BucketingSampler
from .cut_pairs import CutPairsSampler
from .data_source import streaming_shuffle
from .single_cut import SingleCutSampler
from .utils import find_pessimistic_batches
from .zip import ZipSampler



import logging
from typing import List

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.gigaspeech import GIGASPEECH_PARTS, download_gigaspeech, prepare_gigaspeech
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument('corpus_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
@click.option('--subset', type=click.Choice(('auto',) + GIGASPEECH_PARTS), multiple=True,
              default=['auto'], help='Which parts of Gigaspeech to download (by default XL + DEV + TEST).')
@click.option('-j', '--num-jobs', type=int, default=1,
              help='How many threads to use (can give good speed-ups with slow disks).')
def gigaspeech(
        corpus_dir: Pathlike,
        output_dir: Pathlike,
        subset: List[str],
        num_jobs: int
):
    """Gigaspeech ASR data preparation."""
    logging.basicConfig(level=logging.INFO)
    if 'auto' in subset:
        subset = 'auto'
    prepare_gigaspeech(corpus_dir, output_dir=output_dir, dataset_parts=subset, num_jobs=num_jobs)


@download.command(context_settings=dict(show_default=True))
@click.argument('password', type=str)
@click.argument('target_dir', type=click.Path())
@click.option('--subset', type=click.Choice(('auto',) + GIGASPEECH_PARTS), multiple=True,
              default=['auto'], help='Which parts of Gigaspeech to download (by default XL + DEV + TEST).')
@click.option('--host', type=str, default='tsinghua', help='Which host to download Gigaspeech.')
def gigaspeech(
        password: str,
        target_dir: Pathlike,
        subset: List[str],
        host: str
):
    """Gigaspeech download."""
    # Convert (likely one-element) list with "auto" into a string.
    logging.basicConfig(level=logging.INFO)
    if 'auto' in subset:
        subset = 'auto'
    download_gigaspeech(password, target_dir, dataset_parts=subset, host=host)



import click

from lhotse.bin.modes import prepare
from lhotse.recipes import prepare_callhome_english
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument('audio-dir', type=click.Path(exists=True, file_okay=False))
@click.argument('output-dir', type=click.Path())
@click.option('--rttm-dir', type=click.Path(exists=True, file_okay=False))
@click.option('--absolute-paths', default=False,
              help='Whether to return absolute or relative (to the corpus dir) paths for recordings.')
def callhome_english(
        audio_dir: Pathlike,
        output_dir: Pathlike,
        rttm_dir: Pathlike,
        absolute_paths: bool
):
    """
    CallHome English (LDC2001S97) corpus preparation.

    \b
     This script prepares data for speaker diarization on
     a portion of CALLHOME used in the 2000 NIST speaker recognition evaluation.
     The 2000 NIST SRE is required, and has an LDC catalog number LDC2001S97.

    This data is not available for free - your institution needs to have an LDC subscription.

    The data should be located at AUDIO_DIR.
    Optionally, RTTM_DIR can be provided that has the contents of http://www.openslr.org/resources/10/;
    otherwise, we will download it.

    To actually read the audio, you will need the SPH2PIPE binary: you can provide its path,
    so that we will add it in the manifests (otherwise you might need to modify your PATH
    environment variable to find sph2pipe).
    """
    prepare_callhome_english(
        audio_dir=audio_dir,
        rttm_dir=rttm_dir,
        output_dir=output_dir,
        absolute_paths=absolute_paths
    )



"""
The data preparation recipe for the AMI Meeting Corpus.

NOTE on data splits and references:

- The official AMI documentation (http://groups.inf.ed.ac.uk/ami/corpus/datasets.shtml) recommends
three different data partitions: scenario-only, full-corpus, and full-corpus-asr, based on the
task that the data is used for. We provide an argument `partition` which specifies which
partition is to be used.

- We use the latest version of the official annotations: ami_public_manual_1.6.2. This differs from
the Kaldi s5 and s5b recipes which use 1.6.1 (known to have alignment and annotation issues). We
get word-level annotations with time-marks and combine adjacent words into one segment if: (i) they
belong to the same speaker, and (ii) there is no pause between the words. (These supervisions can
later be modified to get larger super-segments based on the task)

NOTE on mic settings: AMI comes with 4 different microphone settings:

- ihm (individual headset microphone)
- sdm (single distant microphone)
- ihm-mix (mix-headset sum)
- mdm (multiple distant microphone)

These can be specified using the `mic` argument.
"""

import html
import itertools
import logging
import urllib.request
import xml.etree.ElementTree as ET
import zipfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Union

from tqdm.auto import tqdm

from lhotse import validate_recordings_and_supervisions
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, Seconds, urlretrieve_progress

# fmt: off
MEETINGS = {
    'EN2001': ['EN2001a', 'EN2001b', 'EN2001d', 'EN2001e'],
    'EN2002': ['EN2002a', 'EN2002b', 'EN2002c', 'EN2002d'],
    'EN2003': ['EN2003a'],
    'EN2004': ['EN2004a'],
    'EN2005': ['EN2005a'],
    'EN2006': ['EN2006a','EN2006b'],
    'EN2009': ['EN2009b','EN2009c','EN2009d'],
    'ES2002': ['ES2002a','ES2002b','ES2002c','ES2002d'],
    'ES2003': ['ES2003a','ES2003b','ES2003c','ES2003d'],
    'ES2004': ['ES2004a','ES2004b','ES2004c','ES2004d'],
    'ES2005': ['ES2005a','ES2005b','ES2005c','ES2005d'],
    'ES2006': ['ES2006a','ES2006b','ES2006c','ES2006d'],
    'ES2007': ['ES2007a','ES2007b','ES2007c','ES2007d'],
    'ES2008': ['ES2008a','ES2008b','ES2008c','ES2008d'],
    'ES2009': ['ES2009a','ES2009b','ES2009c','ES2009d'],
    'ES2010': ['ES2010a','ES2010b','ES2010c','ES2010d'],
    'ES2011': ['ES2011a','ES2011b','ES2011c','ES2011d'],
    'ES2012': ['ES2012a','ES2012b','ES2012c','ES2012d'],
    'ES2013': ['ES2013a','ES2013b','ES2013c','ES2013d'],
    'ES2014': ['ES2014a','ES2014b','ES2014c','ES2014d'],
    'ES2015': ['ES2015a','ES2015b','ES2015c','ES2015d'],
    'ES2016': ['ES2016a','ES2016b','ES2016c','ES2016d'],
    'IB4001': ['IB4001'],
    'IB4002': ['IB4002'],
    'IB4003': ['IB4003'],
    'IB4004': ['IB4004'],
    'IB4005': ['IB4005'],
    'IB4010': ['IB4010'],
    'IB4011': ['IB4011'],
    'IN1001': ['IN1001'],
    'IN1002': ['IN1002'],
    'IN1005': ['IN1005'],
    'IN1007': ['IN1007'],
    'IN1008': ['IN1008'],
    'IN1009': ['IN1009'],
    'IN1012': ['IN1012'],
    'IN1013': ['IN1013'],
    'IN1014': ['IN1014'],
    'IN1016': ['IN1016'],
    'IS1000': ['IS1000a','IS1000b','IS1000c','IS1000d'],
    'IS1001': ['IS1001a','IS1001b','IS1001c','IS1001d'],
    'IS1002': ['IS1002b','IS1002c','IS1002d'],
    'IS1003': ['IS1003a','IS1003b','IS1003c','IS1003d'],
    'IS1004': ['IS1004a','IS1004b','IS1004c','IS1004d'],
    'IS1005': ['IS1005a','IS1005b','IS1005c'],
    'IS1006': ['IS1006a','IS1006b','IS1006c','IS1006d'],
    'IS1007': ['IS1007a','IS1007b','IS1007c','IS1007d'],
    'IS1008': ['IS1008a','IS1008b','IS1008c','IS1008d'],
    'IS1009': ['IS1009a','IS1009b','IS1009c','IS1009d'],
    'TS3003': ['TS3003a','TS3003b','TS3003c','TS3003d'],
    'TS3004': ['TS3004a','TS3004b','TS3004c','TS3004d'],
    'TS3005': ['TS3005a','TS3005b','TS3005c','TS3005d'],
    'TS3006': ['TS3006a','TS3006b','TS3006c','TS3006d'],
    'TS3007': ['TS3007a','TS3007b','TS3007c','TS3007d'],
    'TS3008': ['TS3008a','TS3008b','TS3008c','TS3008d'],
    'TS3009': ['TS3009a','TS3009b','TS3009c','TS3009d'],
    'TS3010': ['TS3010a','TS3010b','TS3010c','TS3010d'],
    'TS3011': ['TS3011a','TS3011b','TS3011c','TS3011d'],
    'TS3012': ['TS3012a','TS3012b','TS3012c','TS3012d'],
}

PARTITIONS = {
    'scenario-only': {
        'train': [meeting for session in [
                'ES2002','ES2005','ES2006','ES2007','ES2008','ES2009','ES2010','ES2012','ES2013',
                'ES2015','ES2016','IS1000','IS1001','IS1002','IS1003','IS1004','IS1005','IS1006',
                'IS1007','TS3005','TS3008','TS3009','TS3010','TS3011','TS3012'
            ] for meeting in MEETINGS[session] if meeting not in ['IS1002a','IS1005d']],
        'dev': [meeting for session in [
                'ES2003','ES2011','IS1008','TS3004','TS3006'
            ] for meeting in MEETINGS[session]],
        'test': [meeting for session in [
                'ES2004','ES2014','IS1009','TS3003','TS3007'
            ] for meeting in MEETINGS[session]]
    },
    'full-corpus': {
        'train': [meeting for session in [
                'ES2002','ES2005','ES2006','ES2007','ES2008','ES2009','ES2010','ES2012','ES2013',
                'ES2015','ES2016','IS1000','IS1001','IS1002','IS1003','IS1004','IS1005','IS1006',
                'IS1007','TS3005','TS3008','TS3009','TS3010','TS3011','TS3012','EN2001','EN2003',
                'EN2004','EN2005','EN2006','EN2009','IN1001','IN1002','IN1005','IN1007','IN1008',
                'IN1009','IN1012','IN1013','IN1014','IN1016'
            ] for meeting in MEETINGS[session]],
        'dev': [meeting for session in [
                'ES2003','ES2011','IS1008','TS3004','TS3006','IB4001','IB4002','IB4003','IB4004',
                'IB4010','IB4011'
            ] for meeting in MEETINGS[session]],
        'test': [meeting for session in [
                'ES2004','ES2014','IS1009','TS3003','TS3007','EN2002'
            ] for meeting in MEETINGS[session]]
    },
    'full-corpus-asr': {
        'train': [meeting for session in [
                'ES2002','ES2005','ES2006','ES2007','ES2008','ES2009','ES2010','ES2012','ES2013',
                'ES2015','ES2016','IS1000','IS1001','IS1002','IS1003','IS1004','IS1005','IS1006',
                'IS1007','TS3005','TS3008','TS3009','TS3010','TS3011','TS3012','EN2001','EN2003',
                'EN2004','EN2005','EN2006','EN2009','IN1001','IN1002','IN1005','IN1007','IN1008',
                'IN1009','IN1012','IN1013','IN1014','IN1016','ES2014','TS3007','ES2003','TS3006'
            ] for meeting in MEETINGS[session]],
        'dev': [meeting for session in [
                'ES2011','IS1008','TS3004','IB4001','IB4002','IB4003','IB4004','IB4010','IB4011'
            ] for meeting in MEETINGS[session]],
        'test': [meeting for session in [
                'ES2004','IS1009','TS3003','EN2002'
            ] for meeting in MEETINGS[session]]
    }
}

MICS = ['ihm','ihm-mix','sdm','mdm']
MDM_ARRAYS = ['Array1','Array2']
MDM_CHANNELS = ['01','02','03','04','05','06','07','08']
# fmt: on


def download_audio(
    target_dir: Path,
    force_download: Optional[bool] = False,
    url: Optional[str] = "http://groups.inf.ed.ac.uk/ami",
    mic: Optional[str] = "ihm",
) -> None:
    # Audios
    for item in tqdm(
        itertools.chain.from_iterable(MEETINGS.values()),
        desc="Downloading AMI meetings",
    ):
        if mic == "ihm":
            headset_num = 5 if item in ("EN2001a", "EN2001d", "EN2001e") else 4
            for m in range(headset_num):
                wav_name = f"{item}.Headset-{m}.wav"
                wav_url = f"{url}/AMICorpusMirror/amicorpus/{item}/audio/{wav_name}"
                wav_dir = target_dir / "wav_db" / item / "audio"
                wav_dir.mkdir(parents=True, exist_ok=True)
                wav_path = wav_dir / wav_name
                if force_download or not wav_path.is_file():
                    urlretrieve_progress(
                        wav_url, filename=wav_path, desc=f"Downloading {wav_name}"
                    )
        elif mic == "ihm-mix":
            wav_name = f"{item}.Mix-Headset.wav"
            wav_url = f"{url}/AMICorpusMirror/amicorpus/{item}/audio/{wav_name}"
            wav_dir = target_dir / "wav_db" / item / "audio"
            wav_dir.mkdir(parents=True, exist_ok=True)
            wav_path = wav_dir / wav_name
            if force_download or not wav_path.is_file():
                urlretrieve_progress(
                    wav_url, filename=wav_path, desc=f"Downloading {wav_name}"
                )
        elif mic == "sdm":
            wav_name = f"{item}.Array1-01.wav"
            wav_url = f"{url}/AMICorpusMirror/amicorpus/{item}/audio/{wav_name}"
            wav_dir = target_dir / "wav_db" / item / "audio"
            wav_dir.mkdir(parents=True, exist_ok=True)
            wav_path = wav_dir / wav_name
            if force_download or not wav_path.is_file():
                urlretrieve_progress(
                    wav_url, filename=wav_path, desc=f"Downloading {wav_name}"
                )
        elif mic == "mdm":
            for array in MDM_ARRAYS:
                for channel in MDM_CHANNELS:
                    wav_name = f"{item}.{array}-{channel}.wav"
                    wav_url = f"{url}/AMICorpusMirror/amicorpus/{item}/audio/{wav_name}"
                    wav_dir = target_dir / "wav_db" / item / "audio"
                    wav_dir.mkdir(parents=True, exist_ok=True)
                    wav_path = wav_dir / wav_name
                    if force_download or not wav_path.is_file():
                        urlretrieve_progress(
                            wav_url, filename=wav_path, desc=f"Downloading {wav_name}"
                        )


def download_ami(
    target_dir: Pathlike = ".",
    annotations_dir: Optional[Pathlike] = None,
    force_download: Optional[bool] = False,
    url: Optional[str] = "http://groups.inf.ed.ac.uk/ami",
    mic: Optional[str] = "ihm",
) -> None:
    """
    Download AMI audio and annotations for provided microphone setting.
    :param target_dir: Pathlike, the path to store the data.
    :param annotations_dir: Pathlike (default = None), path to save annotations zip file
    :param force_download: bool (default = False), if True, download even if file is present.
    :param url: str (default = 'http://groups.inf.ed.ac.uk/ami'), AMI download URL.
    :param mic: str {'ihm','ihm-mix','sdm','mdm'}, type of mic setting.
    """
    target_dir = Path(target_dir)

    annotations_dir = target_dir if not annotations_dir else annotations_dir

    # Audio
    download_audio(target_dir, force_download, url, mic)

    # Annotations
    logging.info("Downloading AMI annotations")
    annotations_name = "annotations.zip"
    annotations_path = annotations_dir / annotations_name
    if annotations_path.exists():
        logging.info(f'Skip downloading annotations as they exist in: {annotations_path}')
        return
    annotations_url = f"{url}/AMICorpusAnnotations/ami_public_manual_1.6.2.zip"
    if force_download or not annotations_path.is_file():
        urllib.request.urlretrieve(annotations_url, filename=annotations_path)


class AmiSegmentAnnotation(NamedTuple):
    text: str
    speaker: str
    gender: str
    begin_time: Seconds
    end_time: Seconds


def parse_ami_annotations(
    annotations_zip: Pathlike, max_pause: float
) -> Dict[str, List[SupervisionSegment]]:
    annotations = defaultdict(dict)
    with zipfile.ZipFile(annotations_zip, "r") as archive:
        # First we get global speaker ids and channels
        global_spk_id = {}
        channel_id = {}
        with archive.open("corpusResources/meetings.xml") as f:
            tree = ET.parse(f)
            for meeting in tree.getroot():
                meet_id = meeting.attrib["observation"]
                for speaker in meeting:
                    local_id = (meet_id, speaker.attrib["nxt_agent"])
                    global_spk_id[local_id] = speaker.attrib["global_name"]
                    channel_id[local_id] = int(speaker.attrib["channel"])

        # Now iterate over all alignments
        for file in archive.namelist():
            if file.startswith("words/") and file[-1] != "/":
                meet_id, x, _, _ = file.split("/")[1].split(".")
                if (meet_id, x) not in global_spk_id:
                    logging.warning(
                        f"No speaker {meet_id}.{x} found! Skipping annotation."
                    )
                    continue
                spk = global_spk_id[(meet_id, x)]
                channel = channel_id[(meet_id, x)]
                tree = ET.parse(archive.open(file))
                key = (meet_id, spk, channel)
                if key not in annotations:
                    annotations[key] = []
                for child in tree.getroot():
                    # If the alignment does not contain start or end time info,
                    # ignore them. Also, only consider words in the alignment XML files.
                    if (
                        "starttime" not in child.attrib
                        or "endtime" not in child.attrib
                        or child.tag != "w"
                    ):
                        continue
                    text = child.text if child.tag == "w" else child.attrib["type"]
                    # to convert HTML escape sequences
                    text = html.unescape(text)
                    annotations[key].append(
                        AmiSegmentAnnotation(
                            text=text,
                            speaker=spk,
                            gender=spk[0],
                            begin_time=float(child.attrib["starttime"]),
                            end_time=float(child.attrib["endtime"]),
                        )
                    )

        # Post-process segments and combine neighboring segments from the same speaker
        for key in annotations:
            new_segs = []
            cur_seg = list(annotations[key])[0]
            for seg in list(annotations[key])[1:]:
                if seg.begin_time - cur_seg.end_time <= max_pause:
                    cur_seg._replace(
                        text=f"{cur_seg.text} {seg.text}", end_time=seg.end_time
                    )
                else:
                    new_segs.append(cur_seg)
                    cur_seg = seg
            new_segs.append(cur_seg)
            annotations[key] = new_segs
    return annotations


# IHM and MDM audio requires grouping multiple channels of AudioSource into
# one Recording.


def prepare_audio_grouped(
    audio_paths: List[Pathlike],
) -> RecordingSet:
    import soundfile as sf

    # Group together multiple channels from the same session.
    # We will use that to create a Recording with multiple sources (channels).
    from cytoolz import groupby

    channel_wavs = groupby(lambda p: p.parts[-3], audio_paths)

    recordings = []
    for session_name, channel_paths in channel_wavs.items():
        audio_sf = sf.SoundFile(str(channel_paths[0]))

        recordings.append(
            Recording(
                id=session_name,
                sources=[
                    AudioSource(type="file", channels=[idx], source=str(audio_path))
                    for idx, audio_path in enumerate(sorted(channel_paths))
                ],
                sampling_rate=audio_sf.samplerate,
                num_samples=audio_sf.frames,
                duration=audio_sf.frames / audio_sf.samplerate,
            )
        )
    return RecordingSet.from_recordings(recordings)


# SDM and IHM-Mix settings do not require any grouping


def prepare_audio_single(
    audio_paths: List[Pathlike],
) -> RecordingSet:
    import soundfile as sf

    recording_manifest = defaultdict(dict)

    recordings = []
    for audio_path in audio_paths:
        session_name = audio_path.parts[-3]
        audio_sf = sf.SoundFile(str(audio_path))
        recordings.append(
            Recording(
                id=session_name,
                sources=[
                    AudioSource(
                        type="file",
                        channels=list(range(audio_sf.channels)),
                        source=str(audio_path),
                    )
                ],
                sampling_rate=audio_sf.samplerate,
                num_samples=audio_sf.frames,
                duration=audio_sf.frames / audio_sf.samplerate,
            )
        )
    return RecordingSet.from_recordings(recordings)


# For IHM mic, each headphone will have its own annotations, while for other mics
# all sources have the same annotation


def prepare_supervision_ihm(
    audio: RecordingSet, annotations: Dict[str, List[AmiSegmentAnnotation]]
) -> SupervisionSet:
    # Create a mapping from a tuple of (session_id, channel) to the list of annotations.
    # This way we can map the supervisions to the right channels in a multi-channel recording.
    annotation_by_id_and_channel = {
        (key[0], key[2]): annotations[key] for key in annotations
    }

    segments = []
    for recording in audio:
        # AMI IHM can have multiple audio sources for each recording
        for source in recording.sources:
            # For each source, "channels" will always be a one-element list
            (channel,) = source.channels
            annotation = annotation_by_id_and_channel.get((recording.id, channel))
            if annotation is None:
                logging.warning(
                    f"No annotation found for recording {recording.id} "
                    f"(file {source.source})"
                )
                continue

            for seg_idx, seg_info in enumerate(annotation):
                duration = seg_info.end_time - seg_info.begin_time
                # Some annotations in IHM setting exceed audio duration, so we
                # ignore such segments
                if seg_info.end_time > recording.duration:
                    logging.warning(
                        f"Segment {recording.id}-{channel}-{seg_idx} exceeds "
                        f"recording duration. Not adding to supervisions."
                    )
                    continue
                if duration > 0:
                    segments.append(
                        SupervisionSegment(
                            id=f"{recording.id}-{channel}-{seg_idx}",
                            recording_id=recording.id,
                            start=seg_info.begin_time,
                            duration=duration,
                            channel=channel,
                            language="English",
                            speaker=seg_info.speaker,
                            gender=seg_info.gender,
                            text=seg_info.text,
                        )
                    )

    return SupervisionSet.from_segments(segments)


def prepare_supervision_other(
    audio: RecordingSet, annotations: Dict[str, List[AmiSegmentAnnotation]]
) -> SupervisionSet:
    annotation_by_id = {(key[0]): annot for key, annot in annotations.items()}

    segments = []
    for recording in audio:
        annotation = annotation_by_id.get(recording.id)
        # In these mic settings, all sources (1 for ihm-mix and sdm and 16 for mdm)
        # will share supervision.
        source = recording.sources[0]
        if annotation is None:
            logging.warning(f"No annotation found for recording {recording.id}")
            continue

        if len(source.channels) > 1:
            logging.warning(
                f"More than 1 channels in recording {recording.id}. "
                f"Creating supervision for channel 0 only."
            )

        for seg_idx, seg_info in enumerate(annotation):
            duration = seg_info.end_time - seg_info.begin_time
            if duration > 0:
                segments.append(
                    SupervisionSegment(
                        id=f"{recording.id}-{seg_idx}",
                        recording_id=recording.id,
                        start=seg_info.begin_time,
                        duration=duration,
                        channel=0,
                        language="English",
                        speaker=seg_info.speaker,
                        gender=seg_info.gender,
                        text=seg_info.text,
                    )
                )
    return SupervisionSet.from_segments(segments)


def prepare_ami(
    data_dir: Pathlike,
    annotations_dir: Optional[Pathlike] = None,
    output_dir: Optional[Pathlike] = None,
    mic: Optional[str] = "ihm",
    partition: Optional[str] = "full-corpus",
    max_pause: Optional[float] = 0.0,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param data_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param mic: str {'ihm','ihm-mix','sdm','mdm'}, type of mic to use.
    :param partition: str {'full-corpus','full-corpus-asr','scenario-only'}, AMI official data split
    :param max_pause: float (default = 0.0), max pause allowed between word segments to combine segments
    :return: a Dict whose key is ('train', 'dev', 'eval'), and the values are dicts of manifests under keys
        'recordings' and 'supervisions'.

    The `partition` and `max_pause` must be chosen depending on the task. For example:
    - Speaker diarization: set `partition="full-corpus"` and `max_pause=0`
    - ASR: set `partition="full-corpus-asr"` and `max_pause=0.3` (or some value in the range 0.2-0.5)
    """
    data_dir = Path(data_dir)
    assert data_dir.is_dir(), f"No such directory: {data_dir}"
    assert mic in MICS, f"Mic {mic} not supported"
    assert partition in PARTITIONS, f"Partition {partition} not supported"

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    logging.info("Parsing AMI annotations")
    annotations_dir = data_dir if not annotations_dir else annotations_dir
    annotations = parse_ami_annotations(
        Path(annotations_dir) / "annotations.zip", max_pause=max_pause
    )

    # Audio
    logging.info("Preparing recording manifests")
    wav_dir = data_dir

    if mic in ["ihm", "mdm"]:
        audio_paths = (
            wav_dir.rglob("*Headset-?.wav")
            if mic == "ihm"
            else wav_dir.rglob("*Array?-0?.wav")
        )
        audio = prepare_audio_grouped(list(audio_paths))
    elif mic in ["ihm-mix", "sdm"]:
        audio_paths = (
            wav_dir.rglob("*Mix-Headset.wav")
            if mic == "ihm-mix"
            else wav_dir.rglob("*Array1-01.wav")
        )
        audio = prepare_audio_single(list(audio_paths))

    # Supervisions
    logging.info("Preparing supervision manifests")
    supervision = (
        prepare_supervision_ihm(audio, annotations)
        if mic == "ihm"
        else prepare_supervision_other(audio, annotations)
    )

    manifests = defaultdict(dict)

    dataset_parts = PARTITIONS[partition]
    for part in ["train", "dev", "test"]:
        # Get recordings for current data split
        audio_part = audio.filter(lambda x: x.id in dataset_parts[part])
        supervision_part = supervision.filter(
            lambda x: x.recording_id in dataset_parts[part]
        )

        # Write to output directory if a path is provided
        if output_dir is not None:
            audio_part.to_json(output_dir / f"recordings_{part}.json")
            supervision_part.to_json(output_dir / f"supervisions_{part}.json")

        validate_recordings_and_supervisions(audio_part, supervision_part)

        # Combine all manifests into one dictionary
        manifests[part] = {"recordings": audio_part, "supervisions": supervision_part}

    return dict(manifests)



"""
BABEL is a collection of corpora created during the IARPA BABEL program:
https://www.iarpa.gov/index.php/research-programs/babel

It has about 25 languages with 40h - 160h of training recordings and ~20h
of development set recordings.
"""
import logging
import re
from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Union

from cytoolz import sliding_window

from lhotse import Recording, RecordingSet, SupervisionSegment, SupervisionSet, validate_recordings_and_supervisions
from lhotse.qa import remove_missing_recordings_and_supervisions, trim_supervisions_to_recordings
from lhotse.utils import Pathlike

BABELCODE2LANG = {
    "101": "Cantonese",
    "102": "Assamese",
    "103": "Bengali",
    "104": "Pashto",
    "105": "Turkish",
    "106": "Tagalog",
    "107": "Vietnamese",
    "201": "Haitian",
    "202": "Swahili",
    "203": "Lao",
    "204": "Tamil",
    "205": "Kurmanji",
    "206": "Zulu",
    "207": "Tok-Pisin",
    "301": "Cebuano",
    "302": "Kazakh",
    "303": "Telugu",
    "304": "Lithuanian",
    "305": "Guarani",
    "306": "Igbo",
    "307": "Amharic",
    "401": "Mongolian",
    "402": "Javanese",
    "403": "Dholuo",
    "404": "Georgian",
}

OOV_PATTERN = re.compile(r'(\(\(\)\)|<foreign>|<prompt>|<overlap>|<hes>)')
SPK_NOISE_PATTERN = re.compile(r'<(limspack|lipsmack|breath|cough)>')
NOISE_PATTERN = re.compile(r'<(click|ring|dtmf|int|sta)>')
SIL_PATTERN = re.compile(r'<no-speech>')
REMOVE_PATTERN = re.compile(r'<(male-to-female|female-to-male)> ')


def prepare_single_babel_language(
        corpus_dir: Pathlike,
        output_dir: Optional[Pathlike] = None,
        no_eval_ok: bool = False
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Prepares manifests using a single BABEL LDC package.

    This function works like the following:

        - first, it will scan `corpus_dir` for a directory named `conversational`;
            if there is more than once, it picks the first one (and emits a warning)
        - then, it will try to find `dev`, `eval`, and `training` splits inside
            (if any of them is not present, it will skip it with a warning)
        - finally, it scans the selected location for SPHERE audio files and transcripts.

    :param corpus_dir: Path to the root of the LDC package with a BABEL language.
    :param output_dir: Path where the manifests are stored.json
    :param no_eval_ok: When set to True, this function won't emit a warning
        that the eval set was not found.
    :return:
    """
    manifests = defaultdict(dict)

    # Auto-detect the location of the "conversational" directory
    orig_corpus_dir = corpus_dir
    corpus_dir = [d for d in corpus_dir.rglob('conversational') if d.is_dir()]
    if not corpus_dir:
        raise ValueError(f"Could not find 'conversational' directory anywhere inside '{orig_corpus_dir}' "
                         f"- please check your path.")
    if len(corpus_dir) > 1:
        # People have very messy data distributions, the best we can do is warn them.
        logging.warning(f"It seems there are multiple 'conversational' directories in '{orig_corpus_dir}' - "
                        f"we are selecting the first one only ({corpus_dir[0]}). Please ensure that you provided "
                        f"the path to a single language's dir, and the root dir for all BABEL languages.")
    corpus_dir = corpus_dir[0].parent

    for split in ('dev', 'eval', 'training'):
        audio_dir = corpus_dir / f'conversational/{split}/audio'
        recordings = RecordingSet.from_recordings(Recording.from_file(p) for p in audio_dir.glob('*.sph'))
        if len(recordings) == 0:
            if split == 'eval' and no_eval_ok:
                continue
            logging.warning(f"No SPHERE files found in {audio_dir}")

        supervisions = []
        text_dir = corpus_dir / f'conversational/{split}/transcription'
        for p in text_dir.glob('*'):
            # p.stem -> BABEL_BP_101_10033_20111024_205740_inLine
            # parts:
            #   0 -> BABEL
            #   1 -> BP
            #   2 -> <language-code> (101)
            #   3 -> <speaker-id> (10033)
            #   4 -> <date> (20111024)
            #   5 -> <hour> (205740)
            #   6 -> channel (inLine) ; inLine <=> A ; outLine <=> B ; "scripted" <=> A
            p0, p1, lang_code, speaker, date, hour, channel, *_ = p.stem.split('_')
            channel = {'inLine': 'A', 'outLine': 'B'}.get(channel, 'A')
            # Fix problematic segments that have two consecutive timestamp lines with no transcript in between
            lines = p.read_text().splitlines() + ['']
            lines = [
                prev_l for prev_l, l in sliding_window(2, lines)
                if not (prev_l.startswith('[') and l.startswith('['))
            ]
            # Add a None at the end so that the last timestamp is only used as "next_timestamp"
            # and ends the iretation (otherwise we'd lose the last segment).
            lines += [None]
            for (timestamp, text), (next_timestamp, _) in sliding_window(2, zip(lines[::2], lines[1::2])):
                try:
                    start = float(timestamp[1:-1])
                    end = float(next_timestamp[1:-1])
                    # Create supervision
                    supervisions.append(
                        SupervisionSegment(
                            id=f'{lang_code}_{speaker}_{channel}_{date}_{hour}_{int(100 * start):06}',
                            recording_id=p.stem,
                            start=start,
                            duration=round(end - start, ndigits=8),
                            channel=0,
                            text=normalize_text(text),
                            language=BABELCODE2LANG[lang_code],
                            speaker=speaker,
                        )
                    )
                except Exception as e:
                    logging.warning(f'Error while parsing segment. Message: {str(e)}')
                    raise ValueError(f"Too many errors while parsing segments (file: '{p}'). "
                                     f"Please check your data or increase the threshold.")
        supervisions = deduplicate_supervisions(supervisions)

        if len(supervisions) == 0:
            logging.warning(f"No supervisions found in {text_dir}")
        supervisions = SupervisionSet.from_segments(supervisions)

        # Fixing and validation of manifests
        if split == 'eval' and len(supervisions) == 0:
            # We won't remove missing recordings for the "eval" split in cases where
            # the user does not have its corresponding transcripts (very likely).
            pass
        else:
            recordings, supervisions = remove_missing_recordings_and_supervisions(recordings, supervisions)
            recordings, supervisions = trim_supervisions_to_recordings(recordings, supervisions)
        validate_recordings_and_supervisions(recordings, supervisions)

        manifests[split] = {
            'recordings': recordings,
            'supervisions': supervisions
        }

        if output_dir is not None:
            output_dir.mkdir(parents=True, exist_ok=True)
            language = BABELCODE2LANG[lang_code]
            save_split = 'train' if split == 'training' else split
            recordings.to_file(output_dir / f'recordings_{language}_{save_split}.json')
            supervisions.to_file(output_dir / f'supervisions_{language}_{save_split}.json')

    return dict(manifests)


def normalize_text(text: str) -> str:
    text = OOV_PATTERN.sub('<unk>', text)
    text = SPK_NOISE_PATTERN.sub('<v-noise>', text)
    text = NOISE_PATTERN.sub('<noise>', text)
    text = SIL_PATTERN.sub('<silence>', text)
    text = REMOVE_PATTERN.sub('', text)
    return text


def deduplicate_supervisions(supervisions: Iterable[SupervisionSegment]) -> List[SupervisionSegment]:
    from cytoolz import groupby
    duplicates = groupby((lambda s: s.id), sorted(supervisions, key=lambda s: s.id))
    filtered = []
    for k, v in duplicates.items():
        if len(v) > 1:
            logging.warning(f'Found {len(v)} supervisions with conflicting IDs ({v[0].id}) '
                            f'- keeping only the first one.')
        filtered.append(v[0])
    return filtered



import warnings
from typing import Optional, Tuple

from lhotse import CutSet, Seconds
from lhotse.dataset.sampling.base import CutSampler, TimeConstraint
from lhotse.dataset.sampling.data_source import DataSource


class CutPairsSampler(CutSampler):
    """
    Samples pairs of cuts from a "source" and "target" CutSet.
    It expects that both CutSet's strictly consist of Cuts with corresponding IDs.
    It behaves like an iterable that yields lists of strings (cut IDs).

    When one of :attr:`max_frames`, :attr:`max_samples`, or :attr:`max_duration` is specified,
    the batch size is dynamic.
    Exactly zero or one of those constraints can be specified.
    Padding required to collate the batch does not contribute to max frames/samples/duration.
    """

    def __init__(
            self,
            source_cuts: CutSet,
            target_cuts: CutSet,
            max_source_frames: int = None,
            max_source_samples: int = None,
            max_source_duration: Seconds = None,
            max_target_frames: int = None,
            max_target_samples: int = None,
            max_target_duration: int = None,
            max_cuts: Optional[int] = None,
            shuffle: bool = False,
            drop_last: bool = False,
            world_size: Optional[int] = None,
            rank: Optional[int] = None,
            seed: int = 0,
    ):
        """
        CutPairsSampler's constructor.

        :param source_cuts: the first ``CutSet`` to sample data from.
        :param target_cuts: the second ``CutSet`` to sample data from.
        :param max_source_frames: The maximum total number of feature frames from ``source_cuts``.
        :param max_source_samples: The maximum total number of audio samples from ``source_cuts``.
        :param max_source_duration: The maximum total recording duration from ``source_cuts``.
        :param max_target_frames: The maximum total number of feature frames from ``target_cuts``.
        :param max_target_samples: The maximum total number of audio samples from ``target_cuts``.
        :param max_target_duration: The maximum total recording duration from ``target_cuts``.
        :param max_cuts: The maximum number of cuts sampled to form a mini-batch.
            By default, this constraint is off.
        :param shuffle: When ``True``, the cuts will be shuffled at the start of iteration.
            Convenient when mini-batch loop is inside an outer epoch-level loop, e.g.:
            `for epoch in range(10): for batch in dataset: ...` as every epoch will see a
            different cuts order.
        :param drop_last: When ``True``, the last batch is dropped if it's incomplete.
        :param world_size: Total number of distributed nodes. We will try to infer it by default.
        :param rank: Index of distributed node. We will try to infer it by default.
        :param seed: Random seed used to consistently shuffle the dataset across different processes.
        """
        super().__init__(
            provide_len=not source_cuts.is_lazy and not target_cuts.is_lazy,
            shuffle=shuffle,
            world_size=world_size,
            rank=rank,
            seed=seed,
        )
        self.source_cuts = DataSource(source_cuts)
        self.target_cuts = DataSource(target_cuts)
        # Constraints
        self.source_constraints = TimeConstraint(
            max_duration=max_source_duration,
            max_samples=max_source_samples,
            max_frames=max_source_frames,
        )
        self.target_constraints = TimeConstraint(
            max_duration=max_target_duration,
            max_samples=max_target_samples,
            max_frames=max_target_frames,
        )
        self.max_cuts = max_cuts
        self.drop_last = drop_last

    @property
    def remaining_duration(self) -> Optional[float]:
        """
        Remaining duration of data left in the sampler (may be inexact due to float arithmetic).
        Not available when the CutSet is read in lazy mode (returns None).

        .. note: For :class:`.CutPairsSampler` we return the source cuts duration.
        """
        return self.source_cuts.remaining_duration

    @property
    def remaining_cuts(self) -> Optional[int]:
        """
        Remaining number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).
        """
        return self.source_cuts.remaining_cuts

    @property
    def num_cuts(self) -> Optional[int]:
        """
        Total number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).
        """
        if self.source_cuts.is_lazy:
            return None
        return len(self.source_cuts)

    def __iter__(self) -> "CutPairsSampler":
        """
        Prepare the dataset for iterating over a new epoch. Will shuffle the data if requested.
        """
        if self.shuffle:
            self.source_cuts.shuffle(self.seed + self.epoch)
            self.target_cuts.shuffle(self.seed + self.epoch)
        iter(self.source_cuts)
        iter(self.target_cuts)
        self.diagnostics.reset()
        return self

    def _next_batch(self) -> Tuple[CutSet, CutSet]:
        # Keep iterating the underlying CutSets as long as we hit or exceed the constraints
        # provided by user (the max number of source_feats or max number of cuts).
        # Note: no actual data is loaded into memory yet because the manifests contain all the metadata
        # required to do this operation.
        self.source_constraints.reset()
        self.target_constraints.reset()
        source_cuts = []
        target_cuts = []
        while True:
            # Check that we have not reached the end of the dataset.
            try:
                # We didn't - grab the next cut
                next_source_cut = next(self.source_cuts)
                next_target_cut = next(self.target_cuts)
                assert next_source_cut.id == next_target_cut.id, (
                    "Sampled source and target cuts with differing IDs. "
                    "Ensure that your source and target cuts have the same length, "
                    "the same IDs, and the same order."
                )
            except StopIteration:
                # No more cuts to sample from: if we have a partial batch,
                # we may output it, unless the user requested to drop it.
                # We also check if the batch is "almost there" to override drop_last.
                if source_cuts and (
                        not self.drop_last
                        or self.source_constraints.close_to_exceeding()
                        or self.target_constraints.close_to_exceeding()
                ):
                    # We have a partial batch and we can return it.
                    assert len(source_cuts) == len(
                        target_cuts
                    ), "Unexpected state: some cuts in source / target are missing their counterparts..."
                    self.diagnostics.keep(source_cuts)
                    return CutSet.from_cuts(source_cuts), CutSet.from_cuts(target_cuts)
                else:
                    # There is nothing more to return or it's discarded:
                    # signal the iteration code to stop.
                    self.diagnostics.discard(source_cuts)
                    raise StopIteration()

            # Check whether the cuts we're about to sample satisfy optional user-requested predicate.
            if self._filter_fn is not None and (
                    not self._filter_fn(next_source_cut)
                    or not self._filter_fn(next_target_cut)
            ):
                # No - try another one.
                continue

            self.source_constraints.add(next_source_cut)
            self.target_constraints.add(next_target_cut)
            next_num_cuts = len(source_cuts) + 1

            # Did we exceed the max_source_frames and max_cuts constraints?
            if (
                    not self.source_constraints.exceeded()
                    and not self.target_constraints.exceeded()
                    and (self.max_cuts is None or next_num_cuts <= self.max_cuts)
            ):
                # No - add the next cut to the batch, and keep trying.
                source_cuts.append(next_source_cut)
                target_cuts.append(next_target_cut)
            else:
                # Yes. Do we have at least one cut in the batch?
                if source_cuts:
                    # Yes. Return it.
                    self.source_cuts.take_back(next_source_cut)
                    self.target_cuts.take_back(next_target_cut)
                    break
                else:
                    # No. We'll warn the user that the constrains might be too tight,
                    # and return the cut anyway.
                    warnings.warn(
                        "The first cut drawn in batch collection violates one of the max_... constraints"
                        "we'll return it anyway. Consider increasing max_source_frames/max_cuts/etc."
                    )
                    source_cuts.append(next_source_cut)
                    target_cuts.append(next_target_cut)

        assert len(source_cuts) == len(
            target_cuts
        ), "Unexpected state: some cuts in source / target are missing their counterparts..."
        self.diagnostics.keep(source_cuts)
        return CutSet.from_cuts(source_cuts), CutSet.from_cuts(target_cuts)



"""
Multilingual LibriSpeech (MLS) dataset is a large multilingual corpus suitable for speech research.
The dataset is derived from read audiobooks from LibriVox and consists of 8 languages -
English, German, Dutch, Spanish, French, Italian, Portuguese, Polish.
It is available at OpenSLR: http://openslr.org/94
"""

import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict

from tqdm.auto import tqdm

from lhotse import *
from lhotse.utils import Pathlike


def prepare_mls(
        corpus_dir: Pathlike,
        output_dir: Optional[Pathlike] = None,
        opus: bool = True,
        num_jobs: int = 1
) -> Dict[str, Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]]:
    """
    Prepare Multilingual LibriSpeech corpus.

    Returns a dict structured like the following:

    .. code-block:: python

        {
            'english': {
                'train': {'recordings': RecordingSet(...), 'supervisions': SupervisionSet(...)},
                'dev': ...,
                'test': ...
            },
            'polish': { ... },
            ...
        }

    :param corpus_dir: Path to the corpus root (directories with specific languages should be inside).
    :param output_dir: Optional path where the manifests should be stored.
    :param opus: Should we scan for OPUS files (otherwise we'll look for FLAC files).
    :param num_jobs: How many jobs should be used for creating recording manifests.
    :return: A dict with structure: ``d[language][split] = {recordings, supervisions}``.
    """
    corpus_dir = Path(corpus_dir)
    output_dir = Path(output_dir) if output_dir is not None else None
    assert corpus_dir.is_dir()

    languages = {
        d.name.split('_')[1]: d
        for d in corpus_dir.glob('mls_*')
        if d.is_dir()
           and '_lm_' not in d.name
           and (opus or not d.name.endswith('opus'))
    }
    logging.info(f'Found MLS languages: {list(languages)}')

    manifests = defaultdict(dict)
    for lang, lang_dir in tqdm(languages.items(), desc='Langauges', total=len(languages)):
        logging.info(f'Processing language: {lang}')

        # Read the speaker to gender mapping.
        spk2gender = {}
        for line in (lang_dir / 'metainfo.txt').read_text().splitlines():
            spk, gender, *_ = line.split('|')
            spk2gender[spk.strip()] = gender.strip()

        for split in tqdm(['test', 'dev', 'train'], desc='Splits'):

            # If everything is ready, read it and skip it.
            recordings_path = None if output_dir is None else output_dir / f'recordings_{lang}_{split}.jsonl.gz'
            supervisions_path = None if output_dir is None else output_dir / f'supervisions_{lang}_{split}.jsonl.gz'
            if (
                    recordings_path is not None and recordings_path.is_file() and
                    supervisions_path is not None and supervisions_path.is_file()
            ):
                logging.info(f'Skipping - {lang}/{split} - already exists!')
                recordings = RecordingSet.from_file(recordings_path)
                supervisions = SupervisionSet.from_file(supervisions_path)
                manifests[lang][split] = {
                    'recordings': recordings,
                    'supervisions': supervisions
                }
                continue

            # Create recordings manifest.
            split_dir = lang_dir / split
            recordings = RecordingSet.from_dir(
                path=split_dir,
                pattern='*.opus' if opus else '*.flac',
                num_jobs=num_jobs,
                force_opus_sampling_rate=16000,
            )

            # Create supervisions manifest.
            supervisions = []
            for line in (split_dir / 'transcripts.txt').read_text().splitlines():
                recording_id, text = line.split('\t')
                speaker = recording_id.split('_')[0]
                supervisions.append(SupervisionSegment(
                    id=recording_id,
                    recording_id=recording_id,
                    text=text,
                    speaker=speaker,
                    gender=spk2gender[speaker],
                    start=0.0,
                    duration=recordings.duration(recording_id),
                    language=lang
                ))
            supervisions = SupervisionSet.from_segments(supervisions)

            # Fix any missing recordings/supervisions.
            recordings, supervisions = fix_manifests(recordings, supervisions)
            validate_recordings_and_supervisions(recordings, supervisions)

            # Save for return.
            manifests[lang][split] = {
                'recordings': recordings,
                'supervisions': supervisions
            }

            # Optional storage on disk.
            if output_dir is not None:
                output_dir.mkdir(exist_ok=True, parents=True)
                recordings.to_jsonl(recordings_path)
                supervisions.to_jsonl(supervisions_path)

    return dict(manifests)



import warnings

with warnings.catch_warnings():
    # Those torchaudio warnings are a real nuisance
    warnings.simplefilter("ignore")
    # noinspection PyUnresolvedReferences
    import torchaudio

from .audio import AudioSource, Recording, RecordingSet
from .augmentation import *
from .cut import MonoCut, CutSet
from .features import *
from .kaldi import load_kaldi_data_dir
from .manipulation import combine, to_manifest, split_parallelize_combine
from .serialization import load_manifest, store_manifest
from .supervision import SupervisionSegment, SupervisionSet
from .tools.env import add_tools_to_path as _add_tools_to_path
from .qa import validate, validate_recordings_and_supervisions, fix_manifests

try:
    # Try to get Lhotse's version (should be created during running pip install / python setup.py ...)
    from .version import __version__
except:
    # Use a default placeholder when the version is unavailable...
    __version__ = '0.0.0.dev'

from . import augmentation
from . import dataset
from . import features
from . import recipes


_add_tools_to_path()



from typing import Any, Mapping


def plot_batch(batch: Mapping[str, Any], supervisions: bool = True, text=True):
    import matplotlib.pyplot as plt

    batch_size = _get_one_of(batch, 'features', 'audio', 'inputs').shape[0]
    fig, axes = plt.subplots(batch_size, figsize=(16, batch_size), sharex=True)

    def _plot_features(key):
        feats = batch[key]
        feat_actors = []
        for idx in range(batch_size):
            feat_actors.append(axes[idx].imshow(feats[idx].numpy().transpose()))
        fig.tight_layout(h_pad=2)
        fig.colorbar(feat_actors[-1], ax=axes)

    if 'features' in batch:
        _plot_features(key='features')

    if 'audio' in batch:
        raise NotImplementedError("Plotting audio for batches is not supported yet.")

    if 'inputs' in batch:
        # For now, assume it is features and not multi-channel audio...
        if len(batch['inputs'].shape) == 3:
            _plot_features(key='inputs')
        else:
            raise NotImplementedError("We could not infer what does the key 'inputs' represent yet.")

    if supervisions and 'supervisions' in batch:
        sups = batch['supervisions']
        for idx in range(len(sups['sequence_idx'])):
            seq_idx = sups['sequence_idx'][idx]
            if all(k in sups for k in ('start_frame', 'num_frames')):
                start, end = sups['start_frame'][idx], sups['start_frame'][idx] + sups['num_frames'][idx]
            elif all(k in sups for k in ('start_sample', 'num_samples')):
                start, end = sups['start_sample'][idx], sups['start_sample'][idx] + sups['num_samples'][idx]
            else:
                raise ValueError(
                    "Cannot plot supervisions: missing 'start_frame/sample' and 'num_frames/samples' fields."
                )
            axes[seq_idx].axvspan(start, end, fill=False, edgecolor='red', linestyle='--', linewidth=4)
            if text and 'text' in sups:
                axes[seq_idx].text(start, -3, sups['text'][idx])


def _get_one_of(d, *keys):
    for k in keys:
        if k in d:
            return d[k]



from pathlib import Path

import click

from lhotse import RecordingSet, SupervisionSet, fix_manifests, load_manifest, validate
from lhotse.bin.modes.cli_base import cli
from lhotse.utils import Pathlike


@cli.command(name='validate')
@click.argument('manifest', type=click.Path(exists=True, dir_okay=False))
@click.option('--read-data/--dont-read-data', default=False,
              help='Should the audio/features data be read from disk to perform additional checks '
                   '(could be extremely slow for large manifests).')
def validate_(manifest: Pathlike, read_data: bool):
    """Validate a Lhotse manifest file."""
    data = load_manifest(manifest)
    validate(data, read_data=read_data)


@cli.command(name='fix')
@click.argument('recordings', type=click.Path(exists=True, dir_okay=False))
@click.argument('supervisions', type=click.Path(exists=True, dir_okay=False))
@click.argument('output_dir', type=click.Path())
def fix_(recordings: Pathlike, supervisions: Pathlike, output_dir: Pathlike):
    """
    Fix a pair of Lhotse RECORDINGS and SUPERVISIONS manifests.
    It removes supervisions without corresponding recordings and vice versa,
    trims the supervisions that exceed the recording, etc.
    Stores the output files in OUTPUT_DIR under the same names as the input
    files.
    """
    output_dir = Path(output_dir)
    recordings = Path(recordings)
    supervisions = Path(supervisions)
    output_dir.mkdir(parents=True, exist_ok=True)
    recs = RecordingSet.from_file(recordings)
    sups = SupervisionSet.from_file(supervisions)
    recs, sups = fix_manifests(recordings=recs, supervisions=sups)
    recs.to_file(output_dir / recordings.name)
    sups.to_file(output_dir / supervisions.name)



import warnings
from dataclasses import dataclass, field
from math import isclose
from typing import Callable, Iterable, Optional, Tuple, Union

from torch import distributed as dist
from torch.utils.data import Sampler

from lhotse import Seconds
from lhotse.cut import Cut
from lhotse.utils import exactly_one_not_null, is_none_or_gt


class CutSampler(Sampler):
    """
    ``CutSampler`` is responsible for collecting batches of cuts, given specified criteria.
    It implements correct handling of distributed sampling in ``DataLoader``,
    so that the cuts are not duplicated across workers.

    Sampling in a ``CutSampler`` is intended to be very quick - it only uses the metadata in
    ``CutSet`` manifest to select the cuts, and is not intended to perform any I/O.

    CutSampler works similarly to PyTorch's DistributedSampler - when :attr:`shuffle=True`,
    you should call ``sampler.set_epoch(epoch)`` at each new epoch to have a different
    ordering of returned elements. However, its actual behaviour is different than that of
    DistributedSampler -- instead of partitioning the underlying cuts into equally sized chunks,
    it will return every N-th batch and skip the other batches (where ``N == world_size``).
    The formula used to determine which batches are returned is:
    ``(batch_idx + (world_size - rank)) % world_size == 0``.
    This ensures that we can return an equal number of batches in all distributed workers
    in spite of using a dynamic batch size, at the cost of skipping at most ``world_size - 1`` batches.

    Example usage::

        >>> dataset = K2SpeechRecognitionDataset(cuts)
        >>> sampler = SingleCutSampler(cuts, max_duration=200, shuffle=True)
        >>> loader = DataLoader(dataset, sampler=sampler, batch_size=None)
        >>> for epoch in range(start_epoch, n_epochs):
        ...     sampler.set_epoch(epoch)
        ...     train(loader)

    .. note::

        For implementers of new samplers:
        Subclasses of CutSampler are expected to implement ``self._next_batch()`` to introduce specific
        sampling logic (e.g. based on filters such as max number of frames/tokens/etc.).
        CutSampler defines ``__iter__()``, which optionally shuffles the cut IDs, and resets
        ``self.cut_idx`` to zero (to be used and incremented inside of ``_next_batch()``.
    """

    def __init__(
            self,
            shuffle: bool = False,
            world_size: Optional[int] = None,
            rank: Optional[int] = None,
            seed: int = 0,
            provide_len: bool = True,
    ) -> None:
        """
        :param shuffle: When ``True``, the cuts will be shuffled at the start of iteration.
            Convenient when mini-batch loop is inside an outer epoch-level loop, e.g.:
            `for epoch in range(10): for batch in dataset: ...` as every epoch will see a
            different cuts order.
        :param world_size: Total number of distributed nodes. We will try to infer it by default.
        :param rank: Index of distributed node. We will try to infer it by default.
        :param seed: Random seed used to consistently shuffle the dataset across different processes.
        :param provide_len: Should we expose the ``__len__`` attribute in this class.
            It makes sense to turn it off when iterating the sampler is somewhat costly for any reason;
            e.g. because the underlying manifest is lazily loaded from the filesystem/somewhere else.
        """
        super().__init__(data_source=None)  # the "data_source" arg is not used in Sampler...
        self.shuffle = shuffle
        self.seed = seed
        self.epoch = 0
        self.provide_len = provide_len

        self._maybe_init_distributed(world_size=world_size, rank=rank)
        self.num_batches = None
        self._filter_fn: Optional[Callable[[Cut], bool]] = None
        self.diagnostics = SamplingDiagnostics()

    def _maybe_init_distributed(self, world_size: Optional[int], rank: Optional[int]):
        if world_size is not None:
            assert world_size >= 1
        if rank is not None:
            assert rank >= 0
        if not dist.is_available() or not dist.is_initialized():
            self.world_size = 1 if world_size is None else world_size
            self.rank = 0 if rank is None else rank
            return
        self.world_size = dist.get_world_size() if world_size is None else world_size
        self.rank = dist.get_rank() if rank is None else rank
        assert self.rank < self.world_size

    def set_epoch(self, epoch: int) -> None:
        """
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        :param epoch: Epoch number.
        """
        self.epoch = epoch

    def filter(self, predicate: Callable[[Cut], bool]) -> None:
        """
        Add a constraint on individual cuts that has to be satisfied to consider them.

        Can be useful when handling large, lazy manifests where it is not feasible to
        pre-filter them before instantiating the sampler.

        When set, we will remove the ``__len__`` attribute on the sampler, as it is now
        determined dynamically.

        Example:
            >>> cuts = CutSet(...)
            ... sampler = SingleCutSampler(cuts, max_duration=100.0)
            ... # Retain only the cuts that have at least 1s and at most 20s duration.
            ... sampler.filter(lambda cut: 1.0 <= cut.duration <= 20.0)
        """
        self._filter_fn = predicate
        self.provide_len = False

    def __iter__(self):
        raise NotImplementedError(
            "Sub-classes of CutSampler have to implement __iter__()"
        )

    def _next_batch(self):
        raise NotImplementedError(
            "Sub-classes of CutSampler have to implement self._next_batch()"
        )

    @property
    def remaining_duration(self) -> Optional[float]:
        """
        Remaining duration of data left in the sampler (may be inexact due to float arithmetic).
        Not available when the CutSet is read in lazy mode (returns None).
        """
        raise NotImplementedError(
            'Sub-classes of CutSampler have to implement self.remaining_duration'
        )

    @property
    def remaining_cuts(self) -> Optional[int]:
        """
        Remaining number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).
        """
        raise NotImplementedError(
            'Sub-classes of CutSampler have to implement self.remaining_cuts'
        )

    @property
    def num_cuts(self) -> Optional[int]:
        """
        Total number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).
        """
        raise NotImplementedError(
            'Sub-classes of CutSampler have to implement self.num_cuts'
        )

    def __len__(self) -> int:
        if not self.provide_len:
            # Fake non-existence of this attribute
            raise TypeError(f"object of type '{type(self).__name__}' has no len()")
        if self.num_batches is None:
            self.num_batches = sum(1 for _ in self)
        return self.num_batches

    def __next__(self):
        # We use the following trick to ensure equal number of batches for each distributed
        # worker:
        # Every time a next batch is required, we will sample self.world_size batches first,
        # and then return the one at position self.rank.
        # This way, if any of the batches raises StopIteration, we'll know to stop early
        # when a given batch was available for one of the nodes, but not for the others.
        batches = []
        for _ in range(self.world_size):
            batches.append(self._next_batch())
        return batches[self.rank]

    def get_report(self) -> str:
        """Returns a string describing the statistics of the sampling process so far."""
        return self.diagnostics.get_report()


@dataclass
class TimeConstraint:
    """
    Represents a time-based constraint for sampler classes.
    It can be defined either as maximum total batch duration (in seconds),
    number of frames, or number of samples.
    These options are mutually exclusive and this class checks for that.

    :class:`TimeConstraint` can be used for tracking whether the criterion has been exceeded
    via the `add(cut)`, `exceeded()` and `reset()` methods.
    It will automatically track the right criterion (i.e. select frames/samples/duration from the cut).
    It can also be a null constraint (never exceeded).
    """

    max_duration: Optional[Seconds] = None
    max_samples: Optional[int] = None
    max_frames: Optional[int] = None
    current: Union[int, Seconds] = 0
    num_cuts: int = 0

    def __post_init__(self) -> None:
        assert exactly_one_not_null(*self._constraints) or all(
            x is None for x in self._constraints
        )
        for c in self._constraints:
            assert is_none_or_gt(c, 0)

    @property
    def _constraints(self) -> Tuple:
        return self.max_duration, self.max_frames, self.max_samples

    def is_active(self) -> bool:
        """Is it an actual constraint, or a dummy one (i.e. never exceeded)."""
        return any(x is not None for x in self._constraints)

    def add(self, cut: Cut) -> None:
        """
        Increment the internal counter for the time constraint,
        selecting the right property from the input ``cut`` object.
        """
        if self.max_frames is not None:
            self.current += cut.num_frames
        if self.max_samples is not None:
            self.current += cut.num_samples
        if self.max_duration is not None:
            self.current += cut.duration
        self.num_cuts += 1

    def exceeded(self) -> bool:
        """Is the constraint exceeded or not."""
        if self.max_frames is not None:
            return self.current > self.max_frames
        if self.max_samples is not None:
            return self.current > self.max_samples
        if self.max_duration is not None:
            return self.current > self.max_duration
        return False

    def close_to_exceeding(self) -> bool:
        """
        Check if the batch is close to satisfying the constraints.
        We define "closeness" as: if we added one more cut that has
        duration/num_frames/num_samples equal to the mean of the current
        batch, then the batch would have exceeded the constraints.
        """
        mean = self.current / self.num_cuts
        if self.max_frames is not None:
            return self.current + mean > self.max_frames
        if self.max_samples is not None:
            return self.current + mean > self.max_samples
        if self.max_duration is not None:
            return self.current + mean > self.max_duration
        return False

    def reset(self) -> None:
        """
        Reset the internal counter (to be used after a batch was created,
        to start collecting a new one).
        """
        self.current = 0
        self.num_cuts = 0

    def __add__(self, other: "TimeConstraint") -> "TimeConstraint":
        for key in ("max_duration", "max_frames", "max_samples"):
            self_attr = getattr(self, key)
            other_attr = getattr(other, key)
            is_none = self_attr is None and other_attr is None
            assert is_none or isclose(self_attr, other_attr), (
                f"To add two TimeConstraint objects, they need to represent the same constraint "
                f"(got self.{key}={self_attr} != other.{key}={other_attr})."
            )
        return TimeConstraint(
            max_duration=self.max_duration,
            max_frames=self.max_frames,
            max_samples=self.max_samples,
            current=self.current + other.current,
            num_cuts=self.num_cuts + other.num_cuts,
        )


@dataclass
class SamplingDiagnostics:
    """
    Utility for collecting diagnostics about the sampling process:
    how many cuts/batches were discarded.
    """

    kept_stats: TimeConstraint = field(
        default_factory=lambda: TimeConstraint(max_duration=float("inf"))
    )
    discarded_stats: TimeConstraint = field(
        default_factory=lambda: TimeConstraint(max_duration=float("inf"))
    )
    num_kept_batches: int = 0
    num_discarded_batches: int = 0

    def keep(self, cuts: Iterable[Cut]) -> None:
        cntr = 0
        for cut in cuts:
            self.kept_stats.add(cut)
            cntr += 1
        if not cntr:
            warnings.warn(
                "Found an accepted batch with zero cuts. This could be an error."
            )
        self.num_kept_batches += 1

    def discard(self, cuts: Iterable[Cut]) -> None:
        cntr = 0
        for cut in cuts:
            self.discarded_stats.add(cut)
            cntr += 1
        if cntr:
            # We don't warn about discarded batches with 0 cuts.
            self.num_discarded_batches += 1

    def reset(self) -> None:
        self.kept_stats.reset()
        self.discarded_stats.reset()
        self.num_kept_batches = 0
        self.num_discarded_batches = 0

    @property
    def total_cuts(self) -> int:
        return self.kept_stats.num_cuts + self.discarded_stats.num_cuts

    @property
    def total_batches(self) -> int:
        return self.num_kept_batches + self.num_discarded_batches

    def get_report(self) -> str:
        """Returns a string describing the statistics of the sampling process so far."""
        if self.total_batches == 0 or self.total_cuts == 0:
            return (
                "Sampling statistics unavailable: the SamplerDiagnostics received no cuts or batches. "
                "If this is unexpected, and you're using a custom sampler, ensure that the sampler "
                "is registering the batches in SamplerDiagnostics."
            )
        return (
            f"Sampling statistics: \n"
            f"Kept {self.kept_stats.num_cuts:d}/{self.total_cuts:d} "
            f"({self.kept_stats.num_cuts / self.total_cuts:.2%}) cuts "
            f"({self.discarded_stats.num_cuts:d} cuts discarded).\n"
            f"Kept {self.num_kept_batches:d}/{self.total_batches:d} "
            f"({self.num_kept_batches / self.total_batches:.2%}) batches "
            f"({self.num_discarded_batches:d} batches discarded).\n"
            f"Overall, {round(self.discarded_stats.current):d} seconds of supervision were discarded."
        )

    def __add__(self, other: "SamplingDiagnostics") -> "SamplingDiagnostics":
        return SamplingDiagnostics(
            kept_stats=self.kept_stats + other.kept_stats,
            discarded_stats=self.discarded_stats + other.discarded_stats,
            num_kept_batches=self.num_kept_batches + other.num_kept_batches,
            num_discarded_batches=self.num_discarded_batches
                                  + other.num_discarded_batches,
        )



from . import cut_transforms
from . import input_strategies
from . import sampling
from . import signal_transforms
from .cut_transforms import *
from .diarization import DiarizationDataset
from .input_strategies import AudioSamples, OnTheFlyFeatures, PrecomputedFeatures
from .sampling import *
from .signal_transforms import GlobalMVN, RandomizedSmoothing, SpecAugment
from .source_separation import (
    DynamicallyMixedSourceSeparationDataset,
    PreMixedSourceSeparationDataset,
    SourceSeparationDataset,
)
from .speech_recognition import K2SpeechRecognitionDataset
from .speech_synthesis import SpeechSynthesisDataset
from .unsupervised import (
    DynamicUnsupervisedDataset,
    UnsupervisedDataset,
    UnsupervisedWaveformDataset,
)
from .vad import VadDataset
from .vis import plot_batch



"""
About the Switchboard corpus

    This is conversational telephone speech collected as 2-channel, 8kHz-sampled
    data.  We are using just the Switchboard-1 Phase 1 training data.
    The catalog number LDC97S62 (Switchboard-1 Release 2) corresponds, we believe,
    to what we have.  We also use the Mississippi State transcriptions, which
    we download separately from
    http://www.isip.piconepress.com/projects/switchboard/releases/switchboard_word_alignments.tar.gz

    This data is not available for free - your institution needs to have an LDC subscription.
"""
import tarfile
from itertools import chain
from pathlib import Path
from typing import Dict, Optional, Union

from lhotse import validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, check_and_rglob, urlretrieve_progress

SWBD_TEXT_URL = 'http://www.isip.piconepress.com/projects/switchboard/releases/switchboard_word_alignments.tar.gz'


def prepare_switchboard(
        audio_dir: Pathlike,
        transcripts_dir: Optional[Pathlike] = None,
        sentiment_dir: Optional[Pathlike] = None,
        output_dir: Optional[Pathlike] = None,
        omit_silence: bool = True,
        absolute_paths: bool = False
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Prepare manifests for the Switchboard corpus.
    We create two manifests: one with recordings, and the other one with text supervisions.
    When ``sentiment_dir`` is provided, we create another supervision manifest with sentiment annotations.

    :param audio_dir: Path to ``LDC97S62`` package.
    :param transcripts_dir: Path to the transcripts directory (typically named "swb_ms98_transcriptions").
        If not provided, the transcripts will be downloaded.
    :param sentiment_dir: Optional path to ``LDC2020T14`` package which contains sentiment annotations
        for SWBD segments.
    :param output_dir: Directory where the manifests should be written. Can be omitted to avoid writing.
    :param omit_silence: Whether supervision segments with ``[silence]`` token should be removed or kept.
    :param absolute_paths: Whether to return absolute or relative (to the corpus dir) paths for recordings.
    :return: A dict with manifests. The keys are: ``{'recordings', 'supervisions'}``.
    """
    if transcripts_dir is None:
        transcripts_dir = download_and_untar()
    audio_paths = check_and_rglob(audio_dir, '*.sph')
    text_paths = check_and_rglob(transcripts_dir, '*trans.text')

    groups = []
    name_to_text = {p.stem.split('-')[0]: p for p in text_paths}
    for ap in audio_paths:
        name = ap.stem.replace('sw0', 'sw')
        groups.append({'audio': ap, 'text-0': name_to_text[f'{name}A'], 'text-1': name_to_text[f'{name}B']})

    recordings = RecordingSet.from_recordings(
        Recording.from_file(group['audio'], relative_path_depth=None if absolute_paths else 3)
        for group in groups
    )
    supervisions = SupervisionSet.from_segments(chain.from_iterable(
        make_segments(
            transcript_path=group[f'text-{channel}'],
            recording=recording,
            channel=channel,
            omit_silence=omit_silence
        )
        for group, recording in zip(groups, recordings)
        for channel in [0, 1]
    ))

    validate_recordings_and_supervisions(recordings, supervisions)

    if sentiment_dir is not None:
        parse_and_add_sentiment_labels(sentiment_dir, supervisions)

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        recordings.to_json(output_dir / 'recordings.json')
        supervisions.to_json(output_dir / 'supervisions.json')
    return {
        'recordings': recordings,
        'supervisions': supervisions
    }


def make_segments(transcript_path: Path, recording: Recording, channel: int, omit_silence: bool = True):
    lines = transcript_path.read_text().splitlines()
    return [
        SupervisionSegment(
            id=segment_id,
            recording_id=recording.id,
            start=float(start),
            duration=round(float(end) - float(start), ndigits=8),
            channel=channel,
            text=' '.join(words),
            language='English',
            speaker=f'{recording.id}A'
        )
        for segment_id, start, end, *words in map(str.split, lines)
        if words[0] != '[silence]' or not omit_silence
    ]


def download_and_untar(
        target_dir: Pathlike = '.',
        force_download: bool = False,
        url: str = SWBD_TEXT_URL
) -> Path:
    target_dir = Path(target_dir)
    transcript_dir = target_dir / 'swb_ms98_transcriptions'
    if transcript_dir.is_dir():
        return transcript_dir
    target_dir.mkdir(parents=True, exist_ok=True)
    tar_name = 'switchboard_word_alignments.tar.gz'
    tar_path = target_dir / tar_name
    if force_download or not tar_path.is_file():
        urlretrieve_progress(url, filename=tar_path, desc=f'Downloading {tar_name}')
    with tarfile.open(tar_path) as tar:
        tar.extractall(path=target_dir)
    return transcript_dir


def parse_and_add_sentiment_labels(sentiment_dir: Pathlike, supervisions: SupervisionSet):
    """Read 'LDC2020T14' sentiment annotations and add then to the supervision segments."""
    import pandas as pd
    # Sanity checks
    sentiment_dir = Path(sentiment_dir)
    labels = sentiment_dir / 'data' / 'sentiment_labels.tsv'
    assert sentiment_dir.is_dir() and labels.is_file()
    # Read the TSV as a dataframe
    df = pd.read_csv(labels, delimiter='\t', names=['id', 'start', 'end', 'sentiment'])
    # We are going to match the segments in LDC2020T14 with the ones we already
    # parsed from ISIP transcripts. We simply look which of the existing segments
    # fall into a sentiment-annotated time span. When doing it this way, we use
    # 51773 out of 52293 available sentiment annotations, which should be okay.
    for _, row in df.iterrows():
        call_id = row['id'].split('_')[0]
        matches = list(supervisions.find(
            recording_id=call_id,
            start_after=row['start'] - 1e-2,
            end_before=row['end'] + 1e-2,
        ))
        if not matches:
            continue
        labels = row['sentiment'].split('#')
        # SupervisionSegments returned from .find() are references to the ones in the
        # SupervisionSet, so we can just modify them. We use the "custom" field
        # to add the sentiment label. Since there were multiple annotators,
        # we add all available labels and leave it up to the user to disambiguate them.
        for segment in matches:
            segment.custom = {f'sentiment{i}': label for i, label in enumerate(labels)}



import click

from lhotse.bin.modes import prepare
from lhotse.recipes.mls import prepare_mls
from lhotse.utils import Pathlike

__all__ = ['mls']


@prepare.command(context_settings=dict(show_default=True))
@click.argument('corpus_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
@click.option('--opus/--flac', type=bool, default=True,
              help='Which codec should be used (OPUS or FLAC)')
@click.option('-j', '--num-jobs', type=int, default=1,
              help='How many threads to use (can give good speed-ups with slow disks).')
def mls(
        corpus_dir: Pathlike,
        output_dir: Pathlike,
        opus: bool,
        num_jobs: int
):
    """
    Multilingual Librispeech (MLS) data preparation.

    Multilingual LibriSpeech (MLS) dataset is a large multilingual corpus suitable for speech research.
    The dataset is derived from read audiobooks from LibriVox and consists of 8 languages -
    English, German, Dutch, Spanish, French, Italian, Portuguese, Polish.
    It is available at OpenSLR: http://openslr.org/94
    """
    prepare_mls(corpus_dir, opus=opus, output_dir=output_dir, num_jobs=num_jobs)
